# File: music.tcl

# Purpose: the Music Window and related commands

#
# Copyright (c) 1997-2001 Tim Baker
#
# This software may be copied and distributed for educational, research, and
# not for profit purposes provided that this copyright and statement are
# included in all such copies.
#

namespace eval NSMusic {

# namespace eval NSMusic
}

# NSMusic::InitModule --
#
#	One-time-only-ever initialization.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::InitModule {} {

	MsgCatInit music

	InitImageIfNeeded Image_Checked checked.gif
	InitImageIfNeeded Image_Unchecked unchecked.gif

	NSObject::New NSMusic

	return
}

# NSMusic::CloseModule --
#
#	One-time-only-ever cleanup.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::CloseModule {} {

	catch {
		destroy [Window music]
	}

	return
}

# NSMusic::NSMusic --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::NSMusic {oop} {

	InitWindow $oop

	set win [Info $oop win]

	Info $oop ignoreSel 0
	Info $oop module,current -1
	Info $oop module,match {}
	Info $oop music,globCount [Global music,globCount]
	Info $oop fontChanged 0
	Info $oop scale,afterId ""

	# Ignore the first call to ScaleCmd() when the scale is created/mapped,
	# otherwise "music position" is called the first time the window is
	# displayed.
	Info $oop ignoreScale 1

	NSWindowManager::RegisterWindow music $win \
		"GetDefaultGeometry $win reqwidth main" "" \
		"NSMusic::DisplayCmd $oop"

	# Update ourself when the listBG value changes
	Info $oop listBG,clientId \
		[NSValueManager::AddClient listBG \
		"NSMusic::ValueChanged_listBG $oop"]

	# Update ourself when the font,knowledge value changes
	Info $oop font,clientId \
		[NSValueManager::AddClient font,knowledge \
		"NSMusic::ValueChanged_font_knowledge $oop"]

	# Destroy the object along with the toplevel (later)
	NSUtils::DestroyObjectWithWidget NSMusic $oop $win

	#
	# Global list of application windows
	#

	Global music,oop $oop
	Window music [Info $oop win]

	return
}

# NSMusic::~NSMusic --
#
#	Object destructor called by NSObject::Delete().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::~NSMusic {oop} {

	NSValueManager::RemoveClient font,knowledge [Info $oop font,clientId]
	NSValueManager::RemoveClient listBG [Info $oop listBG,clientId]

	after cancel [Info $oop scale,afterId]

	return
}

# NSMusic::Info --
#
#	Query and modify info.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::Info {oop info args} {

	global NSMusic

	# Verify the object
	NSObject::CheckObject NSMusic $oop

	# Set info
	if {[llength $args]} {
		switch -- $info {
			default {
				set NSMusic($oop,$info) [lindex $args 0]
			}
		}

	# Get info
	} else {
		switch -- $info {
			default {
				return $NSMusic($oop,$info)
			}
		}
	}

	return
}

# NSMusic::InitWindow --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::InitWindow {oop} {

	set win .music$oop
	toplevel $win
	wm title $win [mc Music]

	wm transient $win [Window main]

	# Do stuff when window closes
	wm protocol $win WM_DELETE_WINDOW "NSMusic::Close $oop"

	# Start out withdrawn (hidden)
	wm withdraw $win

	Info $oop win $win

	InitMenus $oop

	#
	# Divider
	#

	MakeDivider $win.divider1 x

	#
	# Module List
	#

	set font [Value font,knowledge]
	set rowHgt [font metrics $font -linespace]
	if {[image height Image_Checked] > $rowHgt} {
		set rowHgt [image height Image_Checked]
	}
	incr rowHgt 8
	set width 350
	set height 100

	frame $win.frameModule \
		-borderwidth 1 -relief sunken
	set canvistId [NSObject::New NSCanvist $win.frameModule $rowHgt \
		$width $height "NSMusic::NewItemCmd $oop" \
		"NSMusic::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$win.frameModule.scroll set"
	scrollbar $win.frameModule.scroll \
		-command "$canvas yview" -orient vertical

	bind $canvas <Configure> \
		"NSMusic::Configure $oop $canvas"

	NSCanvist::Info $canvistId selectionCmd \
		"NSMusic::SelectionChanged_Module $oop"

	# Do something when a selected group is clicked.
	NSCanvist::Info $canvistId clickCmd \
		"NSMusic::Click_Module $oop"

	Info $oop module,canvistId $canvistId

	grid rowconfigure $win.frameModule 0 -weight 1
	grid columnconfigure $win.frameModule 0 -weight 1
	grid columnconfigure $win.frameModule 1 -weight 0

	grid $canvas \
		-row 0 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $win.frameModule.scroll \
		-row 0 -column 1 -rowspan 1 -columnspan 1 -sticky ns

	#
	# Music position scale
	#

	frame $win.frameConfig \
		-borderwidth 1 -relief sunken

	scale $win.frameModule.scale \
		-orient horizontal -label "Position" \
		-width 8 -sliderlength 12 -length 350 -from 0 -to 100 \
		-showvalue no -command "NSMusic::ScaleCmd $oop" -state disabled

	grid $win.frameModule.scale \
		-row 1 -column 0 -columnspan 2 -sticky ew

	Info $oop scale $win.frameModule.scale

	#
	# Music volume scale
	#

	label $win.frameConfig.label \
		-text Volume:
	scale $win.frameConfig.volume \
		-orient vertical \
		-width 16 -sliderlength 12 -length 100 -from 100 -to 0 \
		-showvalue yes -command "music volume"
	$win.frameConfig.volume set [music volume]

	pack $win.frameConfig.label \
		-side top
	pack $win.frameConfig.volume \
		-side top

	Info $oop volume $win.frameConfig.volume

	#
	# Statusbar
	#

	MakeStatusBar $win.statusBar 20

	#
	# Geometry
	#

	grid columnconfigure $win 0 -weight 1
	grid columnconfigure $win 1 -weight 0
	grid rowconfigure $win 0 -weight 0
	grid rowconfigure $win 1 -weight 1
	grid rowconfigure $win 2 -weight 0
 
	grid $win.divider1 \
		-row 0 -column 0 -rowspan 1 -columnspan 2 -sticky ew
	grid $win.frameModule \
		-row 1 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $win.statusBar \
		-row 2 -column 0 -rowspan 1 -columnspan 2 -sticky ew
	grid $win.frameConfig \
		-row 0 -column 1 -rowspan 2 -columnspan 1 -sticky news

	#
	# KeyPress bindings
	#

	bind $win <KeyPress-Escape> "NSMusic::Close $oop"
	bind $win <Control-KeyPress-w> "NSMusic::Close $oop"

	bind $win <FocusIn> "
		if {\[string equal %W $win]} {
			focus $canvas
		}
	"

	return
}

# NSMusic::InitMenus --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::InitMenus {oop} {

	variable MenuString

	# Default accelerator modifier
	set mod "Ctrl"

	set win [Info $oop win]

	#
	# Menu bar
	#

	Info $oop mbarId [NSObject::New NSMenu $win -tearoff 0 \
		-postcommand "NSMusic::SetupMenus $oop" -identifier MENUBAR]
	set mbar [Info $oop mbarId]

	# Context-sensitive help
	NSMenu::Info $mbar menuSelectCmd "NSMusic::MenuSelect $oop"

	# Call our command when an entry is invoked
	NSMenu::Info $mbar invokeCmd "NSMusic::MenuInvoke $oop"

	#
	# Music Menu
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_MUSIC
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_MUSIC -label [mc Music] -underline 0 -identifier M_MUSIC

	set entries {}
	lappend entries [list -type checkbutton -label [mc "Use Music"] \
		-variable Global(music,play) -identifier E_USE_MUSIC]
	lappend entries [list -type checkbutton -label [mc Randomize] \
		-variable Global(music,random) -identifier E_RANDOMIZE]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc "Set Directory..."] \
		-underline 0 -identifier E_DIRECTORY]
	lappend entries [list -type command -label [mc "Refresh Module List"] \
		-underline 0 -identifier E_UPDATE]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc Close] \
		-underline 0 -accelerator $mod+W -identifier E_CLOSE]

	NSMenu::MenuInsertEntries $mbar -end MENU_MUSIC $entries

	set MenuString(M_MUSIC) \
		"Contains commands for using music."
	set MenuString(E_DIRECTORY) \
		"Changes the directory to look for music files in."
	set MenuString(E_UPDATE) \
		"Redisplays the list of music files."
	set MenuString(E_CLOSE) \
		"Closes the window."

	return
}

# NSMusic::SetupMenus --
#
#	Description
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::SetupMenus {oop mbarId} {

	lappend identList E_USE_MUSIC E_RANDOMIZE E_DIRECTORY E_UPDATE E_CLOSE

	NSMenu::MenuEnable $mbarId $identList

	[Info $oop win].statusBar cover show

	return
}

# NSMusic::MenuSelect --
#
#	Displays a help string associated with a menu entry.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::MenuSelect {oop menuId index ident} {

	variable MenuString

	switch -glob -- $ident {
		{} {
			set desc {}
		}

		E_USE_MUSIC {
			if {[Value music,play]} {
				set desc "Turns off music."
			} else {
				set desc "Turns on music."
			}
		}

		E_RANDOMIZE {
			if {[Value music,random]} {
				set desc "Plays songs in order."
			} else {
				set desc "Chooses a song randomly."
			}
		}
				
		default {
			if {[info exists MenuString($ident)]} {
				set desc $MenuString($ident)
			} else {
				set menu [NSMenu::Info $menuId menu]
				set desc [$menu entrycget $index -label]
			}
		}
	}

	[Info $oop win].statusBar cover set $desc
	if {![string length $desc]} {
		if {$menuId == [Info $oop mbarId]} {
			[Info $oop win].statusBar cover hide
		}
	}

	return
}

# NSMusic::MenuInvoke --
#
#	Called when a menu entry is invoked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::MenuInvoke {oop menuId ident} {

	switch -glob -- $ident {
		E_USE_MUSIC {NSMainWindow::ToggleMusic [Global main,oop]}
		E_RANDOMIZE {ToggleRandomize $oop}
		E_DIRECTORY {MusicDirectory $oop}
		E_UPDATE {SetList_Module $oop}
		E_CLOSE {Close $oop}
	}

	return
}

# NSMusic::DisplayCmd --
#
#	Called by NSWindowManager::Display().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::DisplayCmd {oop message first} {

	switch -- $message {
		preDisplay {
			# Read the music directory if it was never done
			if {![Global music,globCount]} {
				GlobMusicDirectory
			}

			# Select which song is playing since it may change elsewhere.
			SelectCurrentModule $oop
		}
		postDisplay {
			if {$first || [Info $oop fontChanged]} {
				SetList_Module $oop
				Info $oop fontChanged 0

			# Somebody reread the music directory
			} elseif {[Info $oop music,globCount] != [Global music,globCount]} {
				SetList_Module $oop
			}
			SynchScale $oop
		}
		postWithdraw {
			after cancel [Info $oop scale,afterId]
		}
	}

	return
}

# NSMusic::Close --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::Close {oop} {

	NSWindowManager::Undisplay music

	return
}

# NSMusic::Click_Module --
#
#	Do something when a selected module is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::Click_Module {oop canvistId row} {

	return
}

# NSMusic::SelectionChanged_Module --
#
#	Called when the module selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::SelectionChanged_Module {oop canvistId select deselect} {

	set win [Info $oop win]
	set scale [Info $oop scale]

	set selection [NSCanvist::Selection $canvistId]
	set count [llength $selection]

	# Nothing is selected. Clear the assign list.
	if {!$count} {
		Info $oop module,current -1
		return
	}

	set row [lindex $selection 0]
	Info $oop module,current $row

	if {[Info $oop ignoreSel]} return

	set match [Info $oop module,match]
	set module [lindex $match $row]

	PlaySong $module

	return
}

# NSMusic::SelectCurrentModule --
#
#	Select the module that is currently being played.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::SelectCurrentModule {oop} {

	set canvistId [Info $oop module,canvistId]

	set song [Global music,song]
	if {[string length $song]} {
		set index [lsearch -exact [Info $oop module,match] $song]
	} else {
		set index -1
	}

	Info $oop ignoreSel 1
	if {$index == -1} {
		NSCanvist::RemoveSelection $canvistId
	} elseif {$index < [NSCanvist::Info $canvistId count]} {
		NSCanvist::UpdateSelection $canvistId $index all
	}
	Info $oop ignoreSel 0
	
	return
}

# NSMusic::SetList_Module --
#
#	Set the module list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::SetList_Module {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop module,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	set fileList [Global music,fileName]
	set itemList {}
	foreach name $fileList {
		lappend itemList [list "" $name]
	}
	NSCanvist::InsertMany $canvistId end $itemList

	# Map rows to module filenames
	Info $oop module,match $fileList

	Info $oop music,globCount [Global music,globCount]

	# Select which song is playing
	SelectCurrentModule $oop

	return
}

# NSMusic::StatusBar --
#
#	Display text in the status bar, perhaps clearing it later.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::StatusBar {oop text zap} {

	set win [Info $oop win]
	set label [$win.statusBar itemcget t1 -label]
	$label configure -text $text
	if {$zap} {
		NSUtils::ZapLabel $label
	}

	return
}

# NSMusic::NewItemCmd --
#
#	Called by NSCanvist::InsertItem() to create a list row.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::NewItemCmd {oop canvistId y image text} {

	set c [NSCanvist::Info $canvistId canvas]
	set lineHeight [NSCanvist::Info $canvistId rowHgt]
	set font [Value font,knowledge]

	set cw [font measure $font "W"]
	set fw [font measure $font $text]
	set fh [font metrics $font -linespace]
	set diff [expr {int([expr {($lineHeight - $fh) / 2}])}]

	# Selection rectangle inside row
	lappend itemIdList [$c create rectangle 2 [expr {$y + 2}] \
		[expr {([winfo width $c] - 1) - 2}] [expr {$y + $lineHeight - 2}] \
		-fill "" -outline "" -tags {enabled selrect} -width 2.0]

	set textLeft 4
	if {[string length $image]} {
		set iw [image width $image]
		set ih [image height $image]
		set ydiff [expr {int([expr {($lineHeight - $ih) / 2}])}]
		lappend itemIdList [$c create image 4 [expr {$y + $ydiff}] \
			-image $image -anchor nw -tags enabled]
		set textLeft [expr {4 + $iw + 4}]
	}

	lappend itemIdList [$c create text $textLeft [expr {$y + $diff}] \
		-text $text -anchor nw -font $font -fill White -tags enabled]

	return $itemIdList
}

# NSMusic::HighlightItemCmd --
#
#	Called by NSCanvist::Select() to highlight a row.
#
# Arguments:
#	oop					OOP ID. See above.
#	canvistId					OOP ID of NSCanvist object.
#	state					1 or 0 highlight state.
#	args					List of canvas item ids
#
# Results:
#	What happened.

proc NSMusic::HighlightItemCmd {oop canvistId state args} {

	set canvas [NSCanvist::Info $canvistId canvas]
	set itemIdList $args

	set idRect [lindex $itemIdList 0]

	if {[NSUtils::HasFocus $canvas]} {
		set fill [Value listHilite]
	} else {
		set fill [Value listInactive]
	}

	if {$state} {
		$canvas itemconfigure $idRect -outline $fill

	} else {
		$canvas itemconfigure $idRect -outline {}
	}

	return
}

# NSMusic::Configure --
#
#	Called as a <Configure> event script. Resizes the selection rectangles
#	so they fit properly.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::Configure {oop canvas} {

	# Get the width of the canvas
	set canvasWidth [winfo width $canvas]

	foreach itemId [$canvas find withtag selrect] {
		set coords [$canvas coords $itemId]
		set right [expr {($canvasWidth - 1) - 2}]
		set coords [lreplace $coords 2 2 $right]
		eval $canvas coords $itemId $coords
	}

	return
}

# NSMusic::MusicDirectory --
#
#	Puts up a dialog to allow the user to change the current
#	music directory.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::MusicDirectory {oop} {

	set win .setmusicdir
	toplevel $win
	wm title $win [mc dir-title]
	wm transient $win [Info $oop win]

	frame $win.frameSrc \
		-borderwidth 0

	set frame $win.frameSrc
	message $frame.labelSrc \
		-text [mc dir-prompt] -width 350
	entry $frame.entrySrc \
		-width 60
	button $frame.buttonSrc \
		-text [mc Browse...] \
		-command "NSMusic::MusicDirectory_ChooseDir $oop $win $frame.entrySrc"

	set frame $win.frameSrc
	pack $frame \
		-side top -padx 5 -pady 5
	pack $frame.labelSrc \
		-in $frame -side top -padx 5 -anchor w
	pack $frame.entrySrc \
		-in $frame -side left -padx 5
	pack $frame.buttonSrc \
		-in $frame -side left -padx 5

	MakeDivider $win.divider x
	pack $win.divider \
		-side top -fill x -padx 10 -pady 5

	set frame $win.frameButtons
	frame $frame \
		-borderwidth 0
	button $frame.begin \
		-text [mc OK] -width 9 -default active \
		-command "NSMusic::MusicDirectory_Set $oop $win $win.frameSrc.entrySrc"
	button $frame.cancel \
		-text [mc Cancel] -width 9 -command "destroy $win"

	pack $frame \
		-side top -padx 5 -pady 0 -anchor e
	pack $frame.cancel \
		-side right -padx 5 -pady 5
	pack $frame.begin \
		-side right -padx 5 -pady 5

	NSUtils::SetDefaultButton $win $win.frameButtons.begin

	bind $win <KeyPress-Return> \
		"NSUtils::InvokeDefaultButton $win"
	bind $win <KeyPress-Escape> \
		"tkButtonInvoke $frame.cancel"
	
	# Default
	if {[llength [Value music,directory]]} {
		set path [eval file join [Value music,directory]]
		catch {set path [LongName $path]}
		$win.frameSrc.entrySrc insert 0 [file nativename $path]
		$win.frameSrc.entrySrc selection range 0 end
	}

	focus $win.frameSrc.entrySrc

	WindowPosition $win 2 3

	return
}

# NSMusic::MusicDirectory_ChooseDir --
#
#	Puts up a dialog to allow the user to change the current
#	music directory.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::MusicDirectory_ChooseDir {oop parent entry} {

	set initialDir [$entry get]
	if {![file exists $initialDir] || ![file isdirectory $initialDir]} {
		if {[llength [Value music,directory]]} {
			set initialDir [eval file join [Value music,directory]]
		}
	}
	if {![file exists $initialDir] || ![file isdirectory $initialDir]} {
		set initialDir [Path]
	}

	set newDir [tk_chooseDirectory -mustexist yes -parent $parent \
		-initialdir $initialDir]

	if {[string length $newDir]} {
		$entry delete 0 end
		$entry insert 0 [file nativename [LongName $newDir]]
	}

	return
}

# NSMusic::MusicDirectory_Set --
#
#	Tell the binary the location of the sound directory, if it is
#	valid.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::MusicDirectory_Set {oop parent entry} {

	set path [$entry get]

	# Is it a real directory?
	if {![file exists $path] || ![file isdirectory $path]} {
		tk_messageBox -parent $parent -title "Bad Directory" \
		-message "No such directory\n\n    \"$path\"\n\n\
		Try entering a real directory, and see if that helps."
		return
	}

	destroy $parent
	update
	
	# The value-manager saves/restores the value for us
	Value music,directory [file split $path]

	# Read the list of music modules
	GlobMusicDirectory

	# Display list of music modules in the new directory
	SetList_Module $oop

	return
}

# NSMusic::ValueChanged_font_knowledge --
#
#	Called when the font,knowledge value changes.
#	Updates the Assign Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::ValueChanged_font_knowledge {oop} {

	# Calculate row height
	set rowHgt [font metrics [Value font,knowledge] -linespace]
	if {[image height Image_Checked] > $rowHgt} {
		set rowHgt [image height Image_Checked]
	}
	incr rowHgt 8

	# Set row height of group list
	set canvistId [Info $oop module,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	# Set the list later
	Info $oop fontChanged 1

	NSCanvist::DeleteAll [Info $oop module,canvistId]

	return
}

# NSMusic::ValueChanged_listBG --
#
#	Called when the listBG value changes.
#	Updates the Sound Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::ValueChanged_listBG {oop} {

	set color [Value listBG]

	set canvistId [Info $oop module,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background $color

	return
}

# NSMusic::ScaleCmd --
#
#	Called when the value of a scale changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::ScaleCmd {oop position} {

	set scale [Info $oop scale]

	if {[Info $oop ignoreScale]} {
		Info $oop ignoreScale 0
		return
	}
	music position $position

	return
}

# NSMusic::SynchScale --
#
#	Synch the scale to the current position of the song.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::SynchScale {oop} {

	set scale [Info $oop scale]

	if {[music issong]} {
		set position [music position]
		set length [expr {int([music length] - 1)}]
		if {($position != [$scale get]) ||
			($length != [$scale cget -to])} {
			Info $oop ignoreScale 1
			$scale configure -state normal -to $length \
				-label [format [mc "Position %d/%d"] $position $length]
			$scale set $position
		}
	} elseif {[string compare [$scale cget -state] disabled]} {
		$scale configure -state normal -label [mc "(No current song)"] -to 10
		$scale set 0
		$scale configure -state disabled
	}

	Info $oop scale,afterId [after 100 NSMusic::SynchScale $oop]

	return
}

# NSMusic::ToggleRandomize --
#
#	Toggle the "randomize" setting.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSMusic::ToggleRandomize {oop} {

	Value music,random [Global music,random]

	return
}
