# File: sound.tcl

# Purpose: the Sound Window and related commands

#
# Copyright (c) 1997-2001 Tim Baker
#
# This software may be copied and distributed for educational, research, and
# not for profit purposes provided that this copyright and statement are
# included in all such copies.
#

namespace eval NSSound {

	variable Priv

# namespace eval NSSound
}

# NSSound::InitModule --
#
#	One-time-only-ever initialization.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::InitModule {} {

	variable Priv

	MsgCatInit sound

	InitImageIfNeeded Image_Checked checked.gif
	InitImageIfNeeded Image_Unchecked unchecked.gif

	set Priv(hook) {}
	lappend Priv(hook) event Events
	lappend Priv(hook) monster_attack "Monster Attacks"
	lappend Priv(hook) monster_spell "Monster Spells"
	if {[variant OANGBANDTK ZANGBANDTK]} {
		lappend Priv(hook) martial_art "Martial Arts"
	# OANGBANDTK ZANGBANDTK
	}

	NSObject::New NSSound

	return
}

# NSSound::CloseModule --
#
#	One-time-only-ever cleanup.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::CloseModule {} {

	catch {
		destroy [Window sound]
	}

	return
}

# NSSound::NSSound --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::NSSound {oop} {

	InitWindow $oop

	set winMain [Window main]
	set win [Info $oop win]

	Info $oop config ""
	Info $oop config,current -1
	Info $oop dir ""
	Info $oop dir,current -1
	Info $oop sound,ignoreSel 0
	Info $oop group,current -1
	Info $oop group,keyword ""
	Info $oop sound,current -1
	Info $oop sound,match {}
	Info $oop assign,current -1

	Info $oop fontChanged 0

	NSWindowManager::RegisterWindow sound $win \
		"GetDefaultGeometry $win reqwidth main" "" \
		"NSSound::DisplayCmd $oop"

	# Update ourself when the listBG value changes
	Info $oop listBG,clientId \
		[NSValueManager::AddClient listBG \
		"NSSound::ValueChanged_listBG $oop"]

	# Update ourself when the font,knowledge value changes
	Info $oop font,clientId \
		[NSValueManager::AddClient font,knowledge \
		"NSSound::ValueChanged_font_knowledge $oop"]

	# Destroy the object along with the toplevel (later)
	NSUtils::DestroyObjectWithWidget NSSound $oop $win

	#
	# Global list of application windows
	#

	Global sound,oop $oop
	Window sound $win

	return
}

# NSSound::~NSSound --
#
#	Object destructor called by NSObject::Delete().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::~NSSound {oop} {

	NSValueManager::RemoveClient font,knowledge [Info $oop font,clientId]
	NSValueManager::RemoveClient listBG [Info $oop listBG,clientId]

	return
}

# NSSound::Info --
#
#	Query and modify info.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Info {oop info args} {

	global NSSound

	# Verify the object
	NSObject::CheckObject NSSound $oop

	# Set info
	if {[llength $args]} {
		switch -- $info {
			default {
				set NSSound($oop,$info) [lindex $args 0]
			}
		}

	# Get info
	} else {
		switch -- $info {
			default {
				return $NSSound($oop,$info)
			}
		}
	}

	return
}

# NSSound::InitWindow --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::InitWindow {oop} {

	global NSSound
	variable Priv

	set win .sound$oop
	toplevel $win
	wm title $win [mc Sound]

	wm transient $win [Window main]

	# Do stuff when window closes
	wm protocol $win WM_DELETE_WINDOW "NSSound::Close $oop"

	# Start out withdrawn (hidden)
	wm withdraw $win

	Info $oop win $win

	InitMenus $oop

	#
	# Divider
	#

	MakeDivider $win.divider1 x

	#
	# Tabs!
	#

	set tabsId [NSObject::New NSTabs $win]
	foreach {hook label} $Priv(hook) {
		NSTabs::Add $tabsId [mc $label]
	}
	NSTabs::Info $tabsId invokeCmd "NSSound::InvokeTab $oop"
	NSTabs::Info $tabsId active 1
	Info $oop tabsId $tabsId

	set font [Value font,knowledge]
	set cw [font measure [Global font,fixed,normal] "W"]

	#
	# Config List
	#

	set rowHgt [font metrics $font -linespace]
	incr rowHgt 8
	set height [expr {$rowHgt * 3}]

	set frame $win.frameConfig
	frame $frame \
		-borderwidth 1 -relief sunken
	label $frame.label -text "Configuration" -relief raised -anchor w
	set canvistId [NSObject::New NSCanvist $frame $rowHgt 100 $height \
		"NSSound::NewItemCmd $oop" "NSSound::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$frame.yscroll set"
	scrollbar $frame.yscroll \
		-command "$canvas yview" -orient vert

	bind $canvas <Configure> \
		"NSSound::Configure $oop $canvas"

	NSCanvist::Info $canvistId selectionCmd \
		"NSSound::SelectionChanged_Config $oop"

	NSCanvist::Info $canvistId clickCmd \
		"NSSound::Click_Config $oop"

	NSCanvist::Info $canvistId invokeCmd \
		"NSSound::Invoke_Config $oop"

	Info $oop config,canvistId $canvistId

	pack $frame.label -side top -fill x
	pack $frame.yscroll -side right -fill y
	pack $canvas -side left -expand yes -fill both

	#
	# Event List
	#

	set rowHgt [font metrics $font -linespace]
	if {[image height Image_Checked] > $rowHgt} {
		set rowHgt [image height Image_Checked]
	}
	incr rowHgt 8
	set width [CalcGroupListWidth $oop]

	frame $win.frameGroup \
		-borderwidth 1 -relief sunken
	label $win.frameGroup.label -text "Events" -relief raised -anchor w
	set canvistId [NSObject::New NSCanvist $win.frameGroup $rowHgt $width 100 \
		"NSSound::NewItemCmd $oop" "NSSound::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$win.frameGroup.scroll set"
	scrollbar $win.frameGroup.scroll \
		-command "$canvas yview" -orient vert

	bind $canvas <Configure> \
		"NSSound::Configure $oop $canvas"

	NSCanvist::Info $canvistId selectionCmd \
		"NSSound::SelectionChanged_Group $oop"

	# Do something when a selected group is clicked.
	NSCanvist::Info $canvistId clickCmd \
		"NSSound::Click_Group $oop"

	Info $oop group,canvistId $canvistId

	pack $win.frameGroup.label -side top -fill x
	pack $win.frameGroup.scroll -side right -fill y
	pack $canvas -side left -expand yes -fill both

	#
	# Directory List
	#

	set rowHgt [font metrics $font -linespace]
	incr rowHgt 8
	set height [expr {$rowHgt * 3}]

	set frame $win.frameDir
	frame $frame \
		-borderwidth 1 -relief sunken
	label $frame.label -text "Directory" -relief raised -anchor w
	set canvistId [NSObject::New NSCanvist $frame $rowHgt 100 $height \
		"NSSound::NewItemCmd $oop" "NSSound::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$frame.yscroll set"
	scrollbar $frame.yscroll \
		-command "$canvas yview" -orient vert

	bind $canvas <Configure> \
		"NSSound::Configure $oop $canvas"

	NSCanvist::Info $canvistId selectionCmd \
		"NSSound::SelectionChanged_Dir $oop"

	NSCanvist::Info $canvistId clickCmd \
		"NSSound::Click_Dir $oop"

	NSCanvist::Info $canvistId invokeCmd \
		"NSSound::Invoke_Dir $oop"

	Info $oop dir,canvistId $canvistId

	pack $frame.label -side top -fill x
	pack $frame.yscroll -side right -fill y
	pack $canvas -side left -expand yes -fill both

	#
	# Sound List
	#

	set rowHgt [font metrics $font -linespace]
	incr rowHgt 8
	set width 240
	
	frame $win.frameSound \
		-borderwidth 1 -relief sunken
	label $win.frameSound.label -text "Sound Files" -relief raised -anchor w
	set canvistId [NSObject::New NSCanvist $win.frameSound $rowHgt $width 100 \
		"NSSound::NewItemCmd $oop" "NSSound::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$win.frameSound.scroll set"
	scrollbar $win.frameSound.scroll \
		-command "$canvas yview" -orient vert

	bind $canvas <Configure> \
		"NSSound::Configure $oop $canvas"

	NSCanvist::Info $canvistId selectionCmd \
		"NSSound::SelectionChanged_Sound $oop"

	# Do something when a selected sound is clicked.
	NSCanvist::Info $canvistId clickCmd \
		"NSSound::Click_Sound $oop"

	pack $win.frameSound.label -side top -fill x
	pack $win.frameSound.scroll -side right -fill y
	pack $canvas -side left -expand yes -fill both

	Info $oop sound,canvistId $canvistId

	#
	# Divider
	#

	MakeDivider $win.divider2 y

	#
	# Assign List
	#

	set frame $win.frameAssign
	frame $frame \
		-borderwidth 1 -relief sunken
	label $frame.label -text "Assigned Sounds" -relief raised -anchor w
	set canvistId [NSObject::New NSCanvist $frame $rowHgt 100 100 \
		"NSSound::NewItemCmd $oop" "NSSound::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$frame.yscroll set"
	$canvas configure -height [expr {$rowHgt * 4}]
	scrollbar $frame.yscroll \
		-command "$canvas yview" -orient vert

	NSCanvist::Info $canvistId selectionCmd \
		"NSSound::SelectionChanged_Assign $oop"

	# Do something when a selected sound assignment is clicked.
	NSCanvist::Info $canvistId clickCmd \
		"NSSound::Click_SoundAssign $oop"

	bind $canvas <Configure> \
		"NSSound::Configure $oop $canvas"

	pack $frame.label -side top -fill x
	pack $frame.yscroll -side right -fill y
	pack $canvas -side left -expand yes -fill both

	Info $oop assign,canvistId $canvistId

	#
	# Statusbar
	#

	MakeStatusBar $win.statusBar 20

	# Progress bar used to display progress of listing icons
	set label [$win.statusBar itemcget t1 -label]
	set progId [NSObject::New NSProgress2 $label 225 10]
	[NSProgress2::Info $progId frame] configure -borderwidth 0
	Info $oop progId $progId

	# Clicking in label2 displays the icon assigned to the selected
	# member.
	set label2 [$win.statusBar itemcget t2 -label]
	bind $label2 <ButtonPress-1> \
		"NSSound::SynchView $oop"
	NSStatusText::StatusText $label2 $label \
		"Click to view highlighted rows."

	#
	# Geometry
	#

	grid rowconfig $win 0 -weight 0 -minsize 0
	grid rowconfig $win 1 -weight 0 -minsize 0
	grid rowconfig $win 2 -weight 0 -minsize 0
	grid rowconfig $win 3 -weight 1 -minsize 0
	grid rowconfig $win 4 -weight 0 -minsize 0
	grid rowconfig $win 5 -weight 0 -minsize 0
	grid columnconfig $win 0 -weight 0 -minsize 0
	grid columnconfig $win 1 -weight 0 -minsize 0
	grid columnconfig $win 2 -weight 1 -minsize 0
 
	grid $win.divider1 \
		-row 0 -column 0 -rowspan 1 -columnspan 3 -sticky ew
	grid [NSTabs::Info $tabsId canvas] \
		-row 1 -column 0 -rowspan 1 -columnspan 3 -sticky ew
	grid $win.frameConfig \
		-row 2 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $win.frameGroup \
		-row 3 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $win.frameAssign \
		-row 4 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $win.divider2 \
		-row 2 -column 1 -rowspan 3 -columnspan 1 -sticky ns -padx 2
	grid $win.frameDir \
		-row 2 -column 2 -rowspan 1 -columnspan 1 -sticky news
	grid $win.frameSound \
		-row 3 -column 2 -rowspan 2 -columnspan 1 -sticky news
	grid $win.statusBar \
		-row 5 -column 0 -rowspan 1 -columnspan 3 -sticky ew

	#
	# KeyPress bindings
	#

	bind $win <KeyPress-Escape> "NSSound::Close $oop"
	bind $win <Control-KeyPress-w> "NSSound::Close $oop"
	bind $win <KeyPress-plus> "NSSound::InsertSound $oop"
	bind $win <KeyPress-minus> "NSSound::DeleteSound $oop"

	#
	# Synch the scrollbars when window is shown.
	#

#	bind $win.frameGroup.scroll <Map> "NSSound::SynchScrollBars $oop"

	return
}

# NSSound::InitMenus --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::InitMenus {oop} {

	variable MenuString
	variable Priv

	# Default accelerator modifier
	set mod "Ctrl"

	set win [Info $oop win]

	#
	# Menu bar
	#

	Info $oop mbarId [NSObject::New NSMenu $win -tearoff 0 \
		-postcommand "NSSound::MenuSetup $oop" -identifier MENUBAR]
	set mbar [Info $oop mbarId]

	# Context-sensitive help
	NSMenu::Info $mbar menuSelectCmd "NSSound::MenuSelect $oop"

	# Call our command when an entry is invoked
	NSMenu::Info $mbar invokeCmd "NSSound::MenuInvoke $oop"

	#
	# Sound Menu
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_SOUND
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_SOUND -label [mc Sound] -underline 0 -identifier M_SOUND

	set entries {}
	set i 1
	foreach {hook label} $Priv(hook) {
		lappend entries [list -type radiobutton -label [mc $label] \
			-variable NSSound($oop,radio,hook) -value $hook \
			-accelerator $i -identifier E_HOOK_$i]
		bind $win <KeyPress-$i> "NSSound::SetList_Group $oop $hook"
		incr i
	}
	lappend entries [list -type separator]
	lappend entries [list -type checkbutton -label [mc "Use Sound"] \
		-variable ::NSMainWindow([Global main,oop],useSound) \
		-identifier E_USE_SOUND]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc "History"] \
		-identifier E_HISTORY]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc Close] \
		-underline 0 -accelerator $mod+W -identifier E_CLOSE]

	NSMenu::MenuInsertEntries $mbar -end MENU_SOUND $entries

	#
	# Config Menu
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_CONFIG
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_CONFIG -label [mc Config] -underline 0 -identifier M_CONFIG

	set entries {}
	lappend entries [list -type command -label [mc "New Configuration..."] \
		-underline 0 -identifier E_CONFIG_NEW]
	lappend entries [list -type command -label [mc "Add Configuration..."] \
		-underline 0 -identifier E_CONFIG_ADD]
	lappend entries [list -type command -label [mc "Edit Configuration..."] \
		-underline 0 -identifier E_CONFIG_EDIT]
	lappend entries [list -type command -label [mc "Remove Configuration"] \
		-underline 0 -identifier E_CONFIG_REMOVE]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc "Save"] \
		-underline 0 -identifier E_CONFIG_SAVE]
	lappend entries [list -type command -label [mc "Revert"] \
		-underline 0 -identifier E_CONFIG_REVERT]

	NSMenu::MenuInsertEntries $mbar -end MENU_CONFIG $entries

	#
	# Assign Menu
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_ASSIGN
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_ASSIGN -label [mc Assign] -underline 0 -identifier M_ASSIGN

	set entries {}
	lappend entries [list -type command -label [mc "Insert Sound"] \
		-underline 0 -accelerator plus -identifier E_INSERT]
	lappend entries [list -type command -label [mc "Delete Sound"] \
		-underline 0 -accelerator minus -identifier E_DELETE]

	NSMenu::MenuInsertEntries $mbar -end MENU_ASSIGN $entries

	#
	# Directory Menu
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_DIR
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_DIR -label [mc Directory] -underline 0 -identifier M_DIR

	set entries {}
	lappend entries [list -type command -label [mc "Add Directory..."] \
		-underline 0 -identifier E_DIR_ADD]
	lappend entries [list -type command -label [mc "Edit Directory..."] \
		-underline 0 -identifier E_DIR_EDIT]
	lappend entries [list -type command -label [mc "Remove Directory"] \
		-underline 0 -identifier E_DIR_REMOVE]

	NSMenu::MenuInsertEntries $mbar -end MENU_DIR $entries

	set MenuString(M_SOUND) \
		"Contains commands for using sounds."
	set MenuString(E_INSERT) \
		"Adds another sound to the selected event."
	set MenuString(E_DELETE) \
		"Removes selected sounds from the selected event."
	set MenuString(E_CLOSE) \
		"Closes the window."

	return
}

# NSSound::MenuSetup --
#
#	Description
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::MenuSetup {oop mbarId} {

	variable Priv

	if {[Info $oop group,current] != -1} {
		lappend identList E_INSERT
		if {[Info $oop assign,current] != -1} {
			lappend identList E_DELETE
		}
	}

	set i 0
	foreach {hook label} $Priv(hook) {
		lappend identList E_HOOK_[incr i]
	}

	lappend identList E_USE_SOUND E_HISTORY E_CLOSE

	lappend identList E_CONFIG_NEW E_CONFIG_ADD
	if {[Info $oop config,current] != -1} {
		lappend identList E_CONFIG_EDIT E_CONFIG_REMOVE E_CONFIG_SAVE
	}

	lappend identList E_DIR_ADD
	if {[Info $oop dir,current] != -1} {
		lappend identList E_DIR_EDIT E_DIR_REMOVE
	}

	NSMenu::MenuEnable $mbarId $identList

	[Info $oop win].statusBar cover show

	return
}

# NSSound::MenuSelect --
#
#	Displays a help string associated with a menu entry.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::MenuSelect {oop menuId index ident} {

	variable MenuString

	switch -glob -- $ident {
		{} {
			set desc {}
		}

		E_HOOK_* {
			set desc "Displays this page."
		}

		E_USE_SOUND {
			if {[Value use_sound]} {
				set desc "Turns off sounds."
			} else {
				set desc "Turns on sounds."
			}
		}
				
		default {
			if {[info exists MenuString($ident)]} {
				set desc $MenuString($ident)
			} else {
				set menu [NSMenu::Info $menuId menu]
				set desc [$menu entrycget $index -label]
			}
		}
	}

	[Info $oop win].statusBar cover set $desc
	if {![string length $desc]} {
		if {$menuId == [Info $oop mbarId]} {
			[Info $oop win].statusBar cover hide
		}
	}

	return
}

# NSSound::MenuInvoke --
#
#	Called when a menu entry is invoked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::MenuInvoke {oop menuId ident} {

	variable Priv

	switch -glob -- $ident {
		E_INSERT {InsertSound $oop}
		E_DELETE {DeleteSound $oop}
		E_HOOK_* {
			scan $ident "E_HOOK_%d" hookNum
			SetList_Group $oop [lindex $Priv(hook) [expr {($hookNum - 1) * 2}]]
		}
		E_USE_SOUND {NSMainWindow::ToggleSound [Global main,oop]}
		E_HISTORY {History $oop}
		E_CLOSE {Close $oop}

		E_CONFIG_NEW {Config $oop new}
		E_CONFIG_ADD {
			set types {
				{{Sound Config File} {.snd}}
				{{All files} {*}}
			}
			set path [tk_getOpenFile -filetypes $types -parent [Info $oop win] \
				-initialdir [PathTk config]]
			if {$path == ""} return
			if {[file extension $path] != ".snd"} {
				set answer [tk_messageBox -message [mc config-msg-notsnd] \
					-title [mc config-title-notsnd] -icon question \
					-parent [Info $oop win] -type yesno]
				if {$answer == "no"} return
			}
			set row [llength [angband sound config listof]]
			Config::Sound::ConfigHack $path 1
			SetList_Config $oop
			set canvistId [Info $oop config,canvistId]
			NSCanvist::UpdateSelection $canvistId $row all
			SetList_Dir $oop
		}
		E_CONFIG_EDIT {Config $oop edit}
		E_CONFIG_REMOVE {
			set title [angband sound config cget [Info $oop config] -title]
			set msg [format [mc config-msg-remove] $title]
			set answer [tk_messageBox -title [mc config-title-remove] \
				-message $msg -type yesno -icon warning -parent [Info $oop win]]
			if {$answer == "no"} return
			angband sound config delete [Info $oop config]
			SetList_Config $oop
		}
		E_CONFIG_SAVE {
			set config [Info $oop config]
			Config::Sound::WriteConfig $config
		}
		E_CONFIG_REVERT {
			set msg [format [mc config-msg-revert] $title]
			set answer [tk_messageBox -title [mc config-title-revert] \
				-message $msg -type yesno -icon warning -parent [Info $oop win]]
			if {$answer == "no"} return
		}

		E_DIR_ADD {Directory $oop add}
		E_DIR_EDIT {Directory $oop edit}
		E_DIR_REMOVE {
			set path [angband sound dir cget [Info $oop dir] -path]
			set msg [format [mc dir-msg-remove] $path]
			set answer [tk_messageBox -title [mc dir-title-remove] \
				-message $msg -type yesno \
				-icon warning -parent [Info $oop win]]
			if {$answer == "no"} return
			angband sound dir delete [Info $oop dir]
			SetList_Dir $oop
		}
	}

	return
}

# NSSound::DisplayCmd --
#
#	Called by NSWindowManager::Display().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::DisplayCmd {oop message first} {

	variable Priv
	
	switch -- $message {
		preDisplay {
		}
		postDisplay {
			if {$first || [Info $oop fontChanged]} {

				SetList_Config $oop
				SetList_Group $oop event
				SetList_Dir $oop
#				SetList_Sound $oop
				Info $oop fontChanged 0

				# Select first configuration, update event list
				set cfgList [angband sound config listof]
				if {[llength $cfgList]} {
					set canvistId [Info $oop config,canvistId]
					NSCanvist::UpdateSelection $canvistId 0 all
				}

				# Select first directory, set sound list
				set dirList [angband sound dir listof]
				if {[llength $dirList]} {
					set canvistId [Info $oop dir,canvistId]
					NSCanvist::UpdateSelection $canvistId 0 all
				}
			}

			# If the list of sounds changed, set the list again
			set dir [Info $oop dir]
			if {[info exists ::Global(sound,globCount,$dir)]} {
				if {[Info $oop globCount,$dir] != [Global sound,globCount,$dir]} {
					SetList_Sound $oop
				}
			}
		}
	}

	return
}

# NSSound::Close --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Close {oop} {

	NSWindowManager::Undisplay sound

	return
}

# NSSound::SetList_Config --
#
#	Set the config list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SetList_Config {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop config,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	set configList [angband sound config listof]
	
	foreach config $configList {
		set active [angband sound config cget $config -active]
		set title [angband sound config cget $config -title]
		if {$active} {
			set image Image_Checked
		} else {
			set image Image_Unchecked
		}
		NSCanvist::Insert $canvistId end $image $title
	}

	return
}

# NSSound::SetList_Group --
#
#	Set the group list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SetList_Group {oop group} {

	variable Priv

	set win [Info $oop win]
	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	set config [Info $oop config]
	set dir [Info $oop dir]
	Info $oop group $group

	Progress $oop open
	
	# Clear the list
	NSCanvist::DeleteAll $canvistId

	# Map rows to sound indexes
	set match {}
	
	set max [angband sound count $group]
	for {set i 0} {$i < $max} {incr i} {

		if {![angband sound exists $group $i]} continue

		set keyword [angband sound keyword $group $i]
		set desc [mc $group,$keyword]

		set image Image_Unchecked
		if {$config != ""} {
			if {[angband sound config count $config $group $keyword]} {
				set image Image_Checked
	
				if {$::DEBUG} {
					if {[llength [Info $oop sound,match]]} {
						set sndList [angband sound assign $config $group $keyword]
						foreach {dir2 snd2} $sndList {
							if {[string compare $dir $dir2]} continue
							if {[lsearch -exact [Info $oop sound,match] $snd2] == -1} {
								dbwin "Can't find sound \"$snd2\" for $keyword\n"
							}
						}
					}
				}
			}
		}

		NSCanvist::Insert $canvistId end $image $desc

		# Map rows to sound indexes
		lappend match $i

		Progress $oop update $i $max
	}

	# Map rows to sound indexes
	Info $oop group,match $match

	# Radiobutton menu entries
	Info $oop radio,hook $group

	set tabsId [Info $oop tabsId]
	set current [NSTabs::Info $tabsId current]
	set tabId [NSTabs::GetNthId $tabsId [expr {[lsearch -exact $Priv(hook) $group] / 2}]]
	if {$tabId != $current} {
		NSTabs::Smaller $tabsId $current
		NSTabs::Bigger $tabsId $tabId
		NSTabs::Info $tabsId current $tabId
	}

	Progress $oop close

	return
}

# NSSound::SetList_Dir --
#
#	Set the dir list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SetList_Dir {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop dir,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	set dirList [angband sound dir listof]

	set i 0
	foreach dir $dirList {
		set path [angband sound dir cget $dir -path]
		NSCanvist::Insert $canvistId end "" "$i: $path"
		incr i
	}

	return
}

# NSSound::SetList_Sound --
#
#	Set the sound list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SetList_Sound {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop sound,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	# Read the list of sound files if needed
	set dir [Info $oop dir]
	if {![info exists ::Global(sound,globCount,$dir)] ||
		![info exists ::NSSound($oop,globCount,$dir)] ||
		![Info $oop globCount,$dir]} {
		GlobSoundDirectory $dir
	}

	Progress $oop open

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	set fileList [Global sound,fileName,$dir]
	set max [llength $fileList]
	
	# Row zero is "None", for "no sound assigned"
	NSCanvist::Insert $canvistId end "" None

	set i 0
	foreach file $fileList {
		NSCanvist::Insert $canvistId end "" $file
		Progress $oop update $i $max
		incr i
	}

	Info $oop sound,match $fileList
	Info $oop globCount,$dir [Global sound,globCount,$dir]

	Progress $oop close

	return
}

# NSSound::SetList_Assign --
#
#	Display the sounds assigned to the selected keyword(s).
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SetList_Assign {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop assign,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]

	set config [Info $oop config]
	set dir [Info $oop dir]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	set soundIndex 0
	set soundList [angband sound assign $config $group $keyword]
	set soundList2 [CorrectSoundAssign $soundList]
	foreach {dir2 snd2} $soundList {dir3 snd3} $soundList2 {

		# Hack -- If the case changed, re-assign it
		if {[string compare $snd2 $snd3]} {
			angband sound assign $config $group $keyword $soundIndex $dir2 $snd3
			set snd2 $snd3
		}
		if {![string length $snd2]} {
			set snd2 None
		}
		set i [lsearch -exact [angband sound dir listof] $dir2]
		NSCanvist::Insert $canvistId end "" "$i: $snd2"
		incr soundIndex
	}

	return
}

# NSSound::Click_Config --
#
#	Do something when a selected config is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Click_Config {oop canvistId row} {

	# Note: this is called after Toggle, which isn't desired

	return
}

# NSSound::Click_Group --
#
#	Do something when a selected group is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Click_Group {oop canvistId row} {

	set config [Info $oop config]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	# There are no sounds assigned
	if {![angband sound config count $config $group $keyword]} return

	set soundIndex 0
	set current [Info $oop assign,current]
	if {$current != -1} {
		set soundIndex $current
	}
	
	# Get the n'th sound
	set list [angband sound assign $config $group $keyword $soundIndex]
	set dir [lindex $list 0]
	set snd [lindex $list 1]
	
	# Play the sound
	angband sound play $dir $snd

	return
}

# NSSound::Click_Dir --
#
#	Do something when a selected directory is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Click_Dir {oop canvistId row} {
}

# NSSound::Click_Sound --
#
#	Do something when a selected sound is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Click_Sound {oop canvistId row} {

	set dir [Info $oop dir]

	# Skip the "None" sound
	incr row -1
	
	# Get the sound name
	set sound [lindex [Info $oop sound,match] $row]

	# Play the sound
	angband sound play $dir $sound

	return
}

# NSSound::Click_SoundAssign --
#
#	Do something when a selected sound assignment is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Click_SoundAssign {oop canvistId row} {

	set config [Info $oop config]
	set dir [Info $oop dir]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	set list [angband sound assign $config $group $keyword $row]
	set dir2 [lindex $list 0]
	set snd [lindex $list 1]

	# Play the sound
	angband sound play $dir2 $snd

	return
}

# NSSound::Invoke_Config --
#
#	Do something when a configuration is double-clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Invoke_Config {oop canvistId x y} {

	# Find the hit row
	set row [NSCanvist::PointToRow $canvistId $x $y]

	Config $oop edit

	return
}

# NSSound::Invoke_Dir --
#
#	Do something when a directory is double-clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Invoke_Dir {oop canvistId x y} {

	# Find the hit row
	set row [NSCanvist::PointToRow $canvistId $x $y]

	Directory $oop edit

	return
}

# NSSound::SelectionChanged_Config --
#
#	Called when the config selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SelectionChanged_Config {oop canvistId select deselect} {

	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	set rowTags [NSCanvist::Info $canvistId rowTags]

	if {![llength $select]} {
		Info $oop config,current -1
		Info $oop config ""

		# Clear checkboxes in event list
		foreach rowTag $rowTags {
			foreach itemId [$canvas find withtag $rowTag] {
				if {[string equal [$canvas type $itemId] image]} break
			}
			set image Image_Unchecked
			$canvas itemconfigure $itemId -image $image
		}

		# Clear selection in event list (and clear assign list)
		set canvistId [Info $oop group,canvistId]
		NSCanvist::RemoveSelection $canvistId

		return
	}

	set row [lindex $select 0]
	Info $oop config,current $row
	Info $oop config [lindex [angband sound config listof] $row]

	set config [Info $oop config]
	set group [Info $oop group]

	# Update the event list to indicate sound assignments
	foreach event [Info $oop group,match] rowTag $rowTags {
		foreach itemId [$canvas find withtag $rowTag] {
			if {[string equal [$canvas type $itemId] image]} break
		}
		set keyword [angband sound keyword $group $event]
		if {[angband sound config count $config $group $keyword]} {
			set image Image_Checked
		} else {
			set image Image_Unchecked
		}
		$canvas itemconfigure $itemId -image $image
	}

	return
}

# NSSound::SelectionChanged_Group --
#
#	Called when the group selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SelectionChanged_Group {oop canvistId select deselect} {

	set win [Info $oop win]
	set selection [NSCanvist::Selection $canvistId]
	set count [llength $selection]

	# Nothing is selected. Clear the assign list.
	if {!$count} {
		Info $oop group,current -1
		Info $oop group,keyword ""
		if {[NSCanvist::Info [Info $oop assign,canvistId] count]} {
			NSCanvist::DeleteAll [Info $oop assign,canvistId]
		}
		$win.statusBar itemconfigure t2 -text ""
		Info $oop oldLabel2 ""
		return
	}

	set config [Info $oop config]
	if {$config == ""} return

	set group [Info $oop group]
	set match [Info $oop group,match]

	# Only one keyword is selected. Display assigned sounds in the
	# assign list.
	if {$count == 1} {
		set row [lindex $selection 0]
		Info $oop group,current $row
	
		set index [lindex $match $row]
		Info $oop group,keyword [angband sound keyword $group $index]
	
		SetList_Assign $oop
		set canvistId [Info $oop assign,canvistId]
		if {[NSCanvist::Info $canvistId count]} {
			NSCanvist::UpdateSelection $canvistId 0 ""
		}

		$win.statusBar itemconfigure t2 -text [Info $oop group,keyword]

	# More than one item is now selected. The sounds assigned to the
	# group,keyword are assigned to the newly selected keywords.
	# This allows the user to assign the same sounds to a number of
	# keywords just by selecting the new keywords.
	} else {

		set keyword [Info $oop group,keyword]
		set soundList [angband sound assign $config $group $keyword]
		foreach row $select {
			set index [lindex $match $row]
			set keyword [angband sound keyword $group $index]
			foreach {dir snd} [angband sound assign $config $group $keyword] {
				angband sound delete $config $group $keyword 0
			}
			foreach {dir snd} $soundList {
				angband sound insert $config $group $keyword 1000 $dir $snd
			}
			UpdateList_Group $oop $row
		}

		$win.statusBar itemconfigure t2 -text "$count selected"
	}

	return
}

# NSSound::SelectionChanged_Dir --
#
#	Called when the directory selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SelectionChanged_Dir {oop canvistId select deselect} {

	if {![llength $select]} {
		Info $oop dir,current -1
		Info $oop dir ""

		# Clear sound list
		set canvistId [Info $oop sound,canvistId]
		NSCanvist::DeleteAll $canvistId
		return
	}

	set row [lindex $select 0]
	Info $oop dir,current $row
	Info $oop dir [lindex [angband sound dir listof] $row]
	SetList_Sound $oop

	return
}

# NSSound::SelectionChanged_Sound --
#
#	Called when the sound selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SelectionChanged_Sound {oop canvistId select deselect} {

	set dir [Info $oop dir]

	if {![llength $select]} {
		Info $oop sound,current -1
		return
	}

	set row [lindex $select 0]
	Info $oop sound,current $row
	
	if {$row} {
		incr row -1
		set sound [lindex [Info $oop sound,match] $row]
		angband sound play $dir $sound
	} else {
		set sound ""
	}

	# Don't assign the sound, just select and play it
	if {[Info $oop sound,ignoreSel]} return

	set row [Info $oop assign,current]
	if {$row == -1} return

	set config [Info $oop config]
	set dir [Info $oop dir]
	set group [Info $oop group]
	set match [Info $oop group,match]
	set keyword [Info $oop group,keyword]

	# All selected keywords are updated to use the new sound.
	# For this to work, each selected keyword must have the
	# same set of assignments.
	foreach row2 [NSCanvist::Selection [Info $oop group,canvistId]] {
		set index [lindex $match $row2]
		set keyword [angband sound keyword $group $index]
		angband sound assign $config $group $keyword $row $dir $sound
	}

	UpdateList_Assign $oop $row

	return
}

# NSSound::SelectionChanged_Assign --
#
#	Called when the assign selection changes.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SelectionChanged_Assign {oop canvistId select deselect} {

	if {![llength $select]} {
		Info $oop assign,current -1
		if {![llength [NSCanvist::Selection $canvistId]]} {
			NSCanvist::RemoveSelection [Info $oop sound,canvistId]
		}
		return
	}

	set row [lindex $select 0]
	Info $oop assign,current $row

	set config [Info $oop config]
	set dir [Info $oop dir]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	set list [angband sound assign $config $group $keyword $row]
	set dir2 [lindex $list 0]
	set snd [lindex $list 1]

	set canvistId [Info $oop sound,canvistId]

	# Selected sound is in current directory
	if {[string equal $dir $dir2]} {

		set row [lsearch -exact [Info $oop sound,match] $snd]
		incr row

		if {[NSCanvist::IsRowSelected $canvistId $row]} {
			angband sound play $dir2 $snd
		} else {

			# Highlight the sound (and play it if it wasn't highlighted)
			Info $oop sound,ignoreSel 1
			NSCanvist::UpdateSelection $canvistId $row all
			Info $oop sound,ignoreSel 0
		}

		# Scroll the sound into view
		### Don't do this cause it's annoying
#		NSCanvist::See $canvistId $row

	# Play the sound
	} else {
		angband sound play $dir2 $snd
	}

	return
}

# NSSound::Toggle --
#
#	Toggle a checkbox in a list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Toggle {oop canvistId tagOrId} {

	if {$canvistId != [Info $oop config,canvistId]} return

	set canvas [NSCanvist::Info $canvistId canvas]
	set row [NSCanvist::ItemRow $canvistId $tagOrId]
puts "$tagOrId $row"
	set config [lindex [angband sound config listof] $row]
	set active [angband sound config cget $config -active]
	if {$active} {
		set image Image_Unchecked
	} else {
		set image Image_Checked
	}
	set active [expr {!$active}]

	$canvas itemconfigure $tagOrId -image $image
	angband sound config configure $config -active $active

	return
}

# NSSound::SynchScrollBars --
#
#	There is a bug (my bug or in Tk?) which prevents the scroll bars
#	from synchronizing when the window is not mapped. So I bind to
#	the <Map> event and synch the scroll bars here.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SynchScrollBars {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	eval $win.frameGroup.scroll set [$canvas yview]
	$canvas yview moveto 0.0

	set canvistId [Info $oop sound,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	eval $win.frameSound.scroll set [$canvas yview]
	$canvas yview moveto 0.0

	return
}

# NSSound::StatusBar --
#
#	Display text in the status bar, perhaps clearing it later.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::StatusBar {oop text zap} {

	set win [Info $oop win]
	set label [$win.statusBar itemcget t1 -label]
	$label configure -text $text
	if {$zap} {
		NSUtils::ZapLabel $label
	}

	return
}

# NSSound::UpdateList --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::UpdateList {oop list row tag args} {

	set canvistId [Info $oop $list,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	
	set rowTag [lindex [NSCanvist::Info $canvistId rowTags] $row]
	set itemIdList [$canvas find withtag $rowTag]
	set itemId [FindItemByTag $canvas $itemIdList $tag]
	eval $canvas itemconfigure $itemId $args

	return
}

# NSSound::UpdateList_Group --
#
#	Configure the Widget canvas item on the row of a list.
#	When a sound is assigned/unassigned, the event list must
#	be updated.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::UpdateList_Group {oop row} {

	set config [Info $oop config]
	set group [Info $oop group]

	set evt [lindex [Info $oop group,match] $row]
	set keyword [angband sound keyword $group $evt]

	if {[angband sound config count $config $group $keyword]} {
		set image Image_Checked
	} else {
		set image Image_Unchecked
	}

	UpdateList $oop group $row image -image $image

	return
}

# NSSound::UpdateList_Assign --
#
#	Configure the Widget canvas item on the row of a list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::UpdateList_Assign {oop row} {

	set config [Info $oop config]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	set list [angband sound assign $config $group $keyword $row]
	set dir2 [lindex $list 0]
	set snd2 [lindex $list 1]
	set i [lsearch -exact [angband sound dir listof] $dir2]
	if {![string length $snd2]} {set snd2 None}

	UpdateList $oop assign $row text -text "$i: $snd2"

	return
}

# NSSound::NewItemCmd --
#
#	Called by NSCanvist::InsertItem() to create a list row.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::NewItemCmd {oop canvistId y image text} {

	set c [NSCanvist::Info $canvistId canvas]
	set lineHeight [NSCanvist::Info $canvistId rowHgt]
	set font [Value font,knowledge]

	set cw [font measure $font "W"]
	set fw [font measure $font $text]
	set fh [font metrics $font -linespace]
	set diff [expr {int([expr {($lineHeight - $fh) / 2}])}]

	# Selection rectangle inside row
	lappend itemIdList [$c create rectangle 2 [expr {$y + 2}] \
		[expr {([winfo width $c] - 1) - 2}] [expr {$y + $lineHeight - 2}] \
		-fill "" -outline "" -tags {enabled selrect} -width 2.0]

	### Note: NSCanvist::ItemRow required "enabled" tag

	set textLeft 4
	if {[string length $image]} {
		set iw [image width $image]
		set ih [image height $image]
		set ydiff [expr {int([expr {($lineHeight - $ih) / 2}])}]
		lappend itemIdList [$c create image 4 [expr {$y + $ydiff}] \
			-image $image -anchor nw -tags {enabled image}]
		set textLeft [expr {4 + $iw + 4}]

		set id [lindex $itemIdList end]
		$c bind $id <ButtonPress-1> \
			"NSSound::Toggle $oop $canvistId $id ; break"
	}

	lappend itemIdList [$c create text $textLeft [expr {$y + $diff}] \
		-text $text -anchor nw -font $font -fill White -tags {enabled text}]

	return $itemIdList
}

# NSSound::HighlightItemCmd --
#
#	Called by NSCanvist::Select() to highlight a row.
#
# Arguments:
#	oop					OOP ID. See above.
#	canvistId					OOP ID of NSCanvist object.
#	state					1 or 0 highlight state.
#	args					List of canvas item ids
#
# Results:
#	What happened.

proc NSSound::HighlightItemCmd {oop canvistId state args} {

	set canvas [NSCanvist::Info $canvistId canvas]
	set itemIdList $args

	set idRect [lindex $itemIdList 0]

	if {[NSUtils::HasFocus $canvas]} {
		set fill [Value listHilite]
	} else {
		set fill [Value listInactive]
	}

	if {$state} {
		$canvas itemconfigure $idRect -outline $fill

	} else {
		$canvas itemconfigure $idRect -outline {}
	}

	return
}

# NSSound::InvokeTab --
#
#	Called when a tab is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::InvokeTab {oop tabsId tabId} {

	variable Priv
	set index [lsearch -exact [NSTabs::Info $tabsId id] $tabId]
	SetList_Group $oop [lindex $Priv(hook) [expr {$index * 2}]]

	return
}

# NSSound::Configure --
#
#	Called as a <Configure> event script. Resizes the selection rectangles
#	so they fit properly.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Configure {oop canvas} {

	# Get the width of the canvas
	set canvasWidth [winfo width $canvas]

	foreach itemId [$canvas find withtag selrect] {
		set coords [$canvas coords $itemId]
		set right [expr {($canvasWidth - 1) - 2}]
		set coords [lreplace $coords 2 2 $right]
		eval $canvas coords $itemId $coords
	}

	return
}

# NSSound::InsertSound --
#
#	Insert a None sound to the selected keyword.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::InsertSound {oop} {

	if {[Info $oop group,current] == -1} return

	set config [Info $oop config]
	set dir [Info $oop dir]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	set index [Info $oop assign,current]
	if {$index == -1} {
		set index [angband sound config count $config $group $keyword]
	}

	set groupSelection [NSCanvist::Selection [Info $oop group,canvistId]]
	set groupMatch [Info $oop group,match]
	
	# Insert a "None" sound into each selected keyword.
	# For this to work, each selected keyword must have the
	# same set of assignments.
	foreach groupRow $groupSelection {
		set groupIndex [lindex $groupMatch $groupRow]
		set keyword2 [angband sound keyword $group $groupIndex]
		angband sound insert $config $group $keyword2 $index $dir ""
	}

	# Redisplay the list of assignments
	SetList_Assign $oop

	# Highlight and view the assignment
	set canvistId [Info $oop assign,canvistId]
	NSCanvist::UpdateSelection $canvistId $index ""
	NSCanvist::See $canvistId $index

	# If this is the only assignment, then give a visual indication
	# that the keyword now has an assignment.
	if {[angband sound config count $config $group $keyword] == 1} {
		foreach groupRow $groupSelection {
			UpdateList_Group $oop $groupRow
		}
	}

	return
}

# NSSound::DeleteSound --
#
#	Delete a frame from the selected keyword.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::DeleteSound {oop} {

	if {[Info $oop group,current] == -1} return

	set config [Info $oop config]
	set group [Info $oop group]
	set keyword [Info $oop group,keyword]

	set index [Info $oop assign,current]
	if {$index == -1} return

	set groupSelection [NSCanvist::Selection [Info $oop group,canvistId]]
	set groupMatch [Info $oop group,match]
	
	# Delete the assignment from each selected keyword.
	# For this to work, each selected keyword must have the
	# same set of assignments.
	foreach groupRow $groupSelection {
		set groupIndex [lindex $groupMatch $groupRow]
		set keyword2 [angband sound keyword $group $groupIndex]
		angband sound delete $config $group $keyword2 $index
	}

	SetList_Assign $oop

	set count [angband sound config count $config $group $keyword]
	if {$count} {
		if {($index > 0) && ($index >= $count - 1)} {
			incr index -1
		}
		set canvistId [Info $oop assign,canvistId]
		NSCanvist::UpdateSelection $canvistId $index ""
		NSCanvist::See $canvistId $index

	} else {
		foreach groupRow $groupSelection {
			set groupIndex [lindex $groupMatch $groupRow]
			set keyword2 [angband sound keyword $group $groupIndex]
			UpdateList_Group $oop $groupRow
		}
	}

	return
}

# NSSound::Progress --
#
#	Show, update, and hide the progress bar.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Progress {oop action args} {

	set win [Info $oop win]
	set progId [Info $oop progId]
	
	switch -- $action {
		open {
			NSProgress2::Zero $progId
			pack [NSProgress2::Info $progId frame] -pady 1 -fill x
			update idletasks
			Info $oop oldLabel2 [$win.statusBar itemcget t2 -text]
		}
		update {
			set cur [lindex $args 0]
			set max [lindex $args 1]
			set bump [expr {(($max / 20) > 40) ? ($max / 20) : 40}]
			if {$cur && ($cur % $bump) == 0} {
				NSProgress2::SetDoneRatio $progId [expr {$cur / double($max)}]
				$win.statusBar itemconfigure t2 -text "$cur/$max"
				update idletasks
			}
		}
		close {
			NSProgress2::SetDoneRatio $progId 1.0
			update idletasks
			pack forget [NSProgress2::Info $progId frame]
			$win.statusBar itemconfigure t2 -text [Info $oop oldLabel2]
		}
	}

	return
}

# NSSound::Config --
#
#	Puts up a dialog to allow the user to create/edit a configuration.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Config {oop action} {

	set win [Info $oop win].config
	toplevel $win
	wm title $win [mc config-title-$action]
	wm transient $win [Info $oop win]

	message $win.prompt \
		-text [mc config-prompt-$action] -width 350
	pack $win.prompt \
		-side top -padx 5 -pady 5 -anchor w

	#
	# Title
	#

	set frame $win.frameTitle
	frame $frame \
		-borderwidth 0
	label $frame.label \
		-text [mc Title:]
	entry $frame.entry \
		-width 40

	pack $frame \
		-side top -anchor w -padx 5 -pady 5
	pack $frame.label \
		-side top -padx 5 -anchor w
	pack $frame.entry \
		-side left -padx 5

	#
	# File
	#

	set frame $win.frameFile
	frame $frame \
		-borderwidth 0
	label $frame.label \
		-text [mc File:]
	entry $frame.entry \
		-width 20

	pack $frame \
		-side top -anchor w -padx 5 -pady 5
	pack $frame.label \
		-side top -padx 5 -anchor w
	pack $frame.entry \
		-side left -padx 5

	MakeDivider $win.divider x
	pack $win.divider \
		-side top -fill x -padx 10 -pady 5

	#
	# Buttons
	#

	set frame $win.frameButtons
	frame $frame \
		-borderwidth 0
	button $frame.begin \
		-text [mc OK] -width 9 -default active \
		-command "NSSound::Config_Accept $oop"
	button $frame.cancel \
		-text [mc Cancel] -width 9 -command "destroy $win"

	pack $frame \
		-side top -padx 5 -pady 0 -anchor e
	pack $frame.cancel \
		-side right -padx 5 -pady 5
	pack $frame.begin \
		-side right -padx 5 -pady 5

	NSUtils::SetDefaultButton $win $win.frameButtons.begin

	bind $win <KeyPress-Return> \
		"NSUtils::InvokeDefaultButton $win"
	bind $win <KeyPress-Escape> \
		"tkButtonInvoke $frame.cancel"

	Info $oop setcfg,win $win
	Info $oop setcfg,action $action
	Info $oop setcfg,config [Info $oop config]

	if {$action == "edit"} {
		set config [Info $oop config]
		set title [angband sound config cget $config -title]
		$win.frameTitle.entry insert 0 $title
		set file [angband sound config cget $config -file]
		$win.frameFile.entry insert 0 $file
	}
	if {$action == "new"} {
		$win.frameTitle.entry insert 0 Untitled
		$win.frameFile.entry insert 0 untitled.snd
	}
	$win.frameTitle.entry selection range 0 end
	focus $win.frameTitle.entry

	WindowPosition $win 2 3

	return
}

# NSSound::Config_Accept --
#
#	Accept changes made to a configuration.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Config_Accept {oop} {

	set config [Info $oop setcfg,config]
	set win [Info $oop setcfg,win]
	set title [$win.frameTitle.entry get]
	set file [$win.frameFile.entry get]

	# Checks for "" as well
	if {[file extension $file] != ".snd"} {
		tk_messageBox -message [mc config-msg-ext] \
			-title [mc config-title-ext]
		return
	}

	switch -- [Info $oop setcfg,action] {
		edit {
			destroy $win
			angband sound config configure $config -title $title -file $file
			set row [lsearch -exact [angband sound config listof] $config]
			UpdateList $oop config $row text -text $title
		}
		new {
			# Warn if there is already a file with that name
			if {[file exists [PathTk config $file]]} {
				set msg [format [mc config-msg-exist] $file]
				set answer [tk_messageBox -message $msg \
					-title [mc config-title-exist] -type yesno \
					-parent $win -icon question]
				if {$answer == "no"} return
			}
			destroy $win
			set row [llength [angband sound config listof]]			
			set config [angband sound config new -title $title -file $file]
			SetList_Config $oop
			set canvistId [Info $oop config,canvistId]
			NSCanvist::UpdateSelection $canvistId $row all
		}
	}

	return
}

# NSSound::Directory --
#
#	Puts up a dialog to allow the user to add/edit a directory.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Directory {oop action} {

	set win .setsounddir
	toplevel $win
	wm title $win [mc dir-title-$action]
	wm transient $win [Info $oop win]
	
	frame $win.frameSrc \
		-borderwidth 0

	set frame $win.frameSrc
	message $frame.labelSrc \
		-text [mc dir-prompt-$action] -width 350
	entry $frame.entrySrc \
		-width 60
	button $frame.buttonSrc \
		-text [mc Browse...] \
		-command "NSSound::Directory_ChooseDir $oop $win $frame.entrySrc"

	set frame $win.frameSrc
	pack $frame \
		-side top -padx 5 -pady 5
	pack $frame.labelSrc \
		-in $frame -side top -padx 5 -anchor w
	pack $frame.entrySrc \
		-in $frame -side left -padx 5
	pack $frame.buttonSrc \
		-in $frame -side left -padx 5

	MakeDivider $win.divider x
	pack $win.divider \
		-side top -fill x -padx 10 -pady 5

	set frame $win.frameButtons
	frame $frame \
		-borderwidth 0
	button $frame.begin \
		-text [mc OK] -width 9 -default active \
		-command "NSSound::Directory_Accept $oop $win $win.frameSrc.entrySrc"
	button $frame.cancel \
		-text [mc Cancel] -width 9 -command "destroy $win"

	pack $frame \
		-side top -padx 5 -pady 0 -anchor e
	pack $frame.cancel \
		-side right -padx 5 -pady 5
	pack $frame.begin \
		-side right -padx 5 -pady 5

	NSUtils::SetDefaultButton $win $win.frameButtons.begin

	bind $win <KeyPress-Return> \
		"NSUtils::InvokeDefaultButton $win"
	bind $win <KeyPress-Escape> \
		"tkButtonInvoke $frame.cancel"

	Info $oop setdir,action $action
	Info $oop setdir,dir [Info $oop dir]

	if {$action == "edit"} {
		set dir [Info $oop dir]
		set path [angband sound dir cget $dir -path]
		$win.frameSrc.entrySrc insert 0 $path
	}
	$win.frameSrc.entrySrc selection range 0 end
	focus $win.frameSrc.entrySrc

	WindowPosition $win 2 3

	return
}

# NSSound::Directory_ChooseDir --
#
#	Puts up a dialog to allow the user to change the current
#	sound directory.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Directory_ChooseDir {oop parent entry} {

	set initialDir [$entry get]
	if {![file exists $initialDir] || ![file isdirectory $initialDir]} {
		set initialDir [CPath]
	}

	set newDir [tk_chooseDirectory -mustexist yes -parent $parent \
		-initialdir $initialDir]

	if {[string length $newDir]} {
		$entry delete 0 end
		$entry insert 0 [file nativename [LongName $newDir]]
	}

	return
}

# NSSound::Directory_Accept --
#
#	Tell the binary the location of the sound directory, if it is
#	valid.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::Directory_Accept {oop win entry} {

	set path [$entry get]
	if {[file exists $path]} {
		set path [file nativename [LongName $path]]
	}

	# Is it a real directory?
	if {![file exists $path] || ![file isdirectory $path]} {
		set answer [tk_messageBox -parent $win -title "No Such Directory" \
			-message "No such directory\n\n    \"$path\"\n\n\
			Use it anyways?" -type yesno]
		if {$answer == "no"} return
	}

	destroy $win
	update

	switch -- [Info $oop setdir,action] {
		add {
			set row [llength [angband sound dir listof]]			
			angband sound dir new -path $path
			SetList_Dir $oop
			set canvistId [Info $oop dir,canvistId]
			NSCanvist::UpdateSelection $canvistId $row all
		}
		edit {
			set dir [Info $oop setdir,dir]
			angband sound dir configure $dir -path $path
			Global sound,globCount,$dir 0
			Info $oop globCount,$dir 0

			set row [lsearch -exact [angband sound dir listof] $dir]
			UpdateList $oop dir $row text -text "$row: $path"
			SetList_Sound $oop
		}
	}

	return
}

# NSSound::ValueChanged_font_knowledge --
#
#	Called when the font,knowledge value changes.
#	Updates the Assign Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::ValueChanged_font_knowledge {oop} {

	# Calculate row height
	set rowHgt [font metrics [Value font,knowledge] -linespace]
	if {[image height Image_Checked] > $rowHgt} {
		set rowHgt [image height Image_Checked]
	}
	incr rowHgt 8

	# Set row height of group list
	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	# Set the width of the group list
	$canvas configure -width [CalcGroupListWidth $oop]

	###
	set rowHgt [font metrics [Value font,knowledge] -linespace]
	incr rowHgt 8

	# Set row height of sound list
	set canvistId [Info $oop sound,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	# Set row height of assign list
	set canvistId [Info $oop assign,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	# Set the list later
	Info $oop fontChanged 1

	NSCanvist::DeleteAll [Info $oop group,canvistId]
	NSCanvist::DeleteAll [Info $oop sound,canvistId]
	NSCanvist::DeleteAll [Info $oop assign,canvistId]

	return
}

# NSSound::ValueChanged_listBG --
#
#	Called when the listBG value changes.
#	Updates the Sound Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::ValueChanged_listBG {oop} {

	set color [Value listBG]

	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background $color

	set canvistId [Info $oop sound,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background $color

	set canvistId [Info $oop assign,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -background $color

	return
}

# NSSound::CalcGroupListWidth --
#
#	Returns the desired width of the group list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::CalcGroupListWidth {oop} {

	variable Priv

	# Get the desired font
	set font [Value font,knowledge]

	# Minimum desired width of the group list
	set maxWidth 100

	# Calculate the maximum width of the sound names in each group
	foreach {group title} $Priv(hook) {

		# Count the sounds in this group
		set max [angband sound count $group]

		# Check each sound
		for {set i 0} {$i < $max} {incr i} {

			# Require real sound
			if {![angband sound exists $group $i]} continue

			# Get the keyword
			set keyword [angband sound keyword $group $i]

			# Get the description
			set desc [mc $group,$keyword]

			# Calculate the width in pixels
			set width [font measure $font $desc]

			# Remember the maximum width
			if {$width > $maxWidth} {
				set maxWidth $width
			}
		}
	}

	return [expr {$maxWidth + [image width Image_Checked] + 20 + 8}]
}

# NSSound::SynchView --
#
#	Scrolls the selected group, sound, and sound assignment into view.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::SynchView {oop} {

	set row [Info $oop group,current]
	if {$row != -1} {
		NSCanvist::See [Info $oop group,canvistId] $row
	}

	set row [Info $oop sound,current]
	if {$row != -1} {
		NSCanvist::See [Info $oop sound,canvistId] $row
	}

	set row [Info $oop assign,current]
	if {$row != -1} {
		NSCanvist::See [Info $oop assign,canvistId] $row
	}

	return
}

# NSSound::History --
#
#	Display list of recent sound events.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::History {oop} {

	variable Priv

	set win [Info $oop win].history
	toplevel $win
	wm title $win "Event History"
	wm transient $win [Info $oop win]

	listbox $win.listbox -background White -borderwidth 0 \
		-highlightthickness 0 -yscrollcommand "$win.yscroll set" \
		-width 50 -height 15 -font [Value font,knowledge] \
		-foreground White -background [Value listBG]
	scrollbar $win.yscroll -orient vertical -command "$win.listbox yview"

	grid rowconfigure $win 0 -weight 1
	grid columnconfigure $win 0 -weight 1
	grid columnconfigure $win 1 -weight 0
	grid $win.listbox -row 0 -column 0 -sticky news
	grid $win.yscroll -row 0 -column 1 -sticky ns

	bind $win <<ListboxSelect>> \
		"NSSound::History_Select $oop"

	foreach {group label} $Priv(hook) {
		set name($group) [mc $label]
	}
	set name(monster) [mc Monster]

	set dirList [angband sound dir listof]

	set history {}
	set count [angband sound history count]
	for {set i 0} {$i < $count} {incr i} {
		set list [angband sound history get $i]
		set dir [lindex $list 0]
		set snd [lindex $list 1]
		set config [lindex $list 2]
		set group [lindex $list 3]
		set keyword [lindex $list 4]
		set string "$name($group): "
		if {$group == "monster"} {
			scan $keyword mon%d r_idx
			append string \"[angband r_info info $r_idx name]\"
		} else {
			append string \"[mc $group,$keyword]\"
		}
		append string ", [angband sound config cget $config -title]"
		set d [lsearch -exact $dirList $dir]
		append string ", $d: $snd"
		$win.listbox insert end $string

		lappend history $list
	}

	$win.listbox see end

	bind $win <KeyPress-Escape> "destroy $win"

	WindowPosition $win 2 3

	Info $oop history,history $history
	Info $oop history,listbox $win.listbox

	return
}

# NSSound::History_Select --
#
#	Handle selection in History Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSound::History_Select {oop} {

	set listbox [Info $oop history,listbox]
	set selection [$listbox curselection]
	set row [lindex $selection 0]

	set history [lindex [Info $oop history,history] $row]
	set dir [lindex $history 0]
	set snd [lindex $history 1]

	angband sound play $dir $snd

	return
}
