# File: squelch.tcl

# Purpose: the Squelch Window and related commands

#
# Copyright (c) 1997-2001 Tim Baker
#
# This software may be copied and distributed for educational, research, and
# not for profit purposes provided that this copyright and statement are
# included in all such copies.
#

namespace eval NSSquelch {

	variable MenuString
	variable Priv

# namespace eval NSSquelch
}

# NSSquelch::InitModule --
#
#	One-time-only-ever initialization.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::InitModule {} {

	variable Priv

	MsgCatInit know squelch

	InitImageIfNeeded Image_Sound sound.gif

	set Priv(page) {}
	lappend Priv(page) page_armor [mc Armor]
	lappend Priv(page) page_weapon [mc Weapons]
	lappend Priv(page) page_other [mc Other]
	
	set Priv(find,string) ""
	set Priv(find,fromStart) 1

	# Split global k_info groups into our Armor/Weapon/Other groups
	foreach {title findSpec} [Global groups,k_info] {
		switch -glob -- $findSpec {
			*TV_SWORD* -
			*TV_POLEARM* -
			*TV_HAFTED* -
			*TV_BOW* -
			*TV_ARROW* {
				lappend Priv(group,weapon) $title $findSpec
			}
			*TV_SHIELD* -
			*TV_CROWN* -
			*TV_HELM* -
			*TV_GLOVES* -
			*TV_BOOTS* -
			*TV_CLOAK* -
			*TV_DRAG_ARMOR* -
			*TV_HARD_ARMOR* -
			*TV_SOFT_ARMOR* {
				lappend Priv(group,armor) $title $findSpec
			}
			default {
				lappend Priv(group,other) $title $findSpec
			}
		}
	}

	# Create the Squelch Window
	NSObject::New NSSquelch

	return
}

# NSSquelch::CloseModule --
#
#	One-time-only-ever cleanup.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::CloseModule {} {

	catch {
		destroy [Window squelch]
	}

	return
}

# NSSquelch::NSSquelch --
#
#	Object constructor called by NSObject::New().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::NSSquelch {oop} {

	variable Priv

	Info $oop page [lindex $Priv(page) 0]

	# Info about selected group & member for each page
	foreach {page label} $Priv(page) {
		Info $oop group,$page [list 0]
		Info $oop member,$page {}
	}

	InitWindow $oop

	# Currently-selected groups and members
	Info $oop group,current {}
	Info $oop member,current {}

	set win [Info $oop win]

	NSWindowManager::RegisterWindow squelch $win \
		"GetDefaultGeometry $win main main" "NSSquelch::SetupCmd $oop" \
		"NSSquelch::DisplayCmd $oop"		

	# Update ourself when the listBG value changes
	Info $oop listBG,clientId \
		[NSValueManager::AddClient listBG \
		"NSSquelch::ValueChanged_listBG $oop"]

	# Update ourself when the font,knowledge value changes
	Info $oop font,clientId \
		[NSValueManager::AddClient font,knowledge \
		"NSSquelch::ValueChanged_font_knowledge $oop"]

	# Destroy the object along with the toplevel (later)
	NSUtils::DestroyObjectWithWidget NSSquelch $oop $win
	
	bind $win <KeyPress-Escape> "NSSquelch::Close $oop"
	bind $win <Control-KeyPress-w> "NSSquelch::Close $oop"
	bind $win <KeyPress-f> "NSSquelch::Find $oop 0"
	bind $win <KeyPress-g> "NSSquelch::Find $oop 1"

	#
	# Global access
	#

	Window squelch $win
	Global squelch,oop $oop

	return
}

# NSSquelch::~NSSquelch --
#
#	Object destructor called by NSObject::Delete().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::~NSSquelch {oop} {

	NSValueManager::RemoveClient listBG [Info $oop listBG,clientId]
	NSValueManager::RemoveClient font,knowledge [Info $oop font,clientId]

	return
}

# NSSquelch::Info --
#
#	Query and modify info.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::Info {oop info args} {

	global NSSquelch

	# Verify the object
	NSObject::CheckObject NSSquelch $oop

	# Set info
	if {[llength $args]} {
		switch -- $info {
			default {
				set NSSquelch($oop,$info) [lindex $args 0]
			}
		}

	# Get info
	} else {
		switch -- $info {
			groupCanvas {
				set canvistId [Info $oop group,canvistId]
				return [NSCanvist::Info $canvistId canvas]
			}
			memberCanvas {
				set canvistId [Info $oop member,canvistId]
				return [NSCanvist::Info $canvistId canvas]
			}
			default {
				return $NSSquelch($oop,$info)
			}
		}
	}

	return
}

# NSSquelch::InitWindow --
#
#	Create the window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::InitWindow {oop} {

	variable Priv

	set win .squelch$oop
	toplevel $win
	wm title $win [mc Squelch]

	wm transient $win [Window main]

	# Start out withdrawn (hidden)
	wm withdraw $win

	# Do stuff when window closes
	wm protocol $win WM_DELETE_WINDOW "NSSquelch::Close $oop"

	Info $oop win $win

	InitMenus $oop

	#
	# Tabs!
	#

	set tabsId [NSObject::New NSTabs $win]
	foreach {page label} $Priv(page) {
		NSTabs::Add $tabsId $label
	}
	NSTabs::Info $tabsId invokeCmd "NSSquelch::InvokeTab $oop"
	NSTabs::Info $tabsId active 1
	Info $oop tabsId $tabsId

	#
	# Group List
	#

	set font [Value font,knowledge]
	set lineSpace [font metrics $font -linespace]
	set rowHgt $lineSpace
	if {[icon size] > $rowHgt} {
		set rowHgt [icon size]
	}
	incr rowHgt 8
	set width [CalcGroupListWidth $oop]

	frame $win.frameGroup \
		-borderwidth 1 -relief sunken
	set canvistId [NSObject::New NSCanvist $win.frameGroup $rowHgt $width 100 \
		"NSSquelch::NewItemCmd $oop" "NSSquelch::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -height [expr {40 * 5}]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$win.frameGroup.scroll set"
	scrollbar $win.frameGroup.scroll \
		-command "$canvas yview" -orient vertical

	NSUtils::SynchScrollBar $canvas $win.frameGroup.scroll

	# Do something when a group item is selected
	NSCanvist::Info $canvistId selectionCmd \
		"NSSquelch::SelectionChanged_Group $oop"

	pack $win.frameGroup.scroll -side right -fill y
	pack $canvas -side left -expand yes -fill both -anchor nw

	Info $oop group,canvistId $canvistId
	Info $oop group,canvas $canvas

	#
	# Member List
	#

	set width 300

	frame $win.frameMember \
		-borderwidth 1 -relief sunken
	set canvistId [NSObject::New NSCanvist $win.frameMember $rowHgt $width 100 \
		"NSSquelch::NewItemCmd $oop" "NSSquelch::HighlightItemCmd $oop"]
	set canvas [NSCanvist::Info $canvistId canvas]
	$canvas configure -height [expr {40 * 5}]
	$canvas configure -background [Value listBG]
	$canvas configure -yscrollcommand "$win.frameMember.scroll set"
	scrollbar $win.frameMember.scroll \
		-command "$canvas yview" -orient vertical

	NSUtils::SynchScrollBar $canvas $win.frameMember.scroll

	# Do something when a member item is selected.
	NSCanvist::Info $canvistId selectionCmd \
		"NSSquelch::SelectionChanged_Member $oop"

	bind $canvas <Configure> \
		"NSSquelch::Configure $oop $canvas"

	Info $oop member,canvistId $canvistId
	Info $oop member,canvas $canvas

	pack $win.frameMember.scroll -side right -fill y
	pack $canvas -side left -expand yes -fill both -anchor nw

	#
	# Statusbar
	#

	MakeStatusBar $win.statusBar 20

	#
	# Geometry
	#

	grid rowconfig $win 0 -weight 0 -minsize 0
	grid rowconfig $win 1 -weight 0 -minsize 0
	grid rowconfig $win 2 -weight 1 -minsize 0
	grid rowconfig $win 3 -weight 0 -minsize 0
	grid columnconfig $win 0 -weight 0 -minsize 0
	grid columnconfig $win 1 -weight 1 -minsize 0
 
	if {[Platform windows]} {
		grid [MakeDivider $win.divider2 x] \
			-row 0 -column 0 -rowspan 1 -columnspan 2 -sticky ew -pady 2
	}
	grid [NSTabs::Info $tabsId canvas] \
		-row 1 -column 0 -rowspan 1 -columnspan 2 -sticky ew
	grid $win.frameGroup \
		-row 2 -column 0 -rowspan 1 -columnspan 1 -sticky ns
	grid $win.frameMember \
		-row 2 -column 1 -rowspan 1 -columnspan 1 -sticky news
	grid $win.statusBar -in $win \
		-row 3 -column 0 -rowspan 1 -columnspan 2 -sticky ew

	return
}

# NSSquelch::InitMenus --
#
#	Create the menus.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::InitMenus {oop} {

	variable MenuString
	variable Priv

	set win [Info $oop win]
	set mod "Ctrl"

	#
	# Menu bar
	#

	set mbar [NSObject::New NSMenu $win -tearoff 0 \
		-postcommand "NSSquelch::SetupMenus $oop" -identifier MENUBAR]
	Info $oop mbarId $mbar

	# Context-sensitive help
	NSMenu::Info $mbar menuSelectCmd "NSSquelch::MenuSelect $oop"

	# Call our command when an entry is invoked
	NSMenu::Info $mbar invokeCmd "NSSquelch::MenuInvoke $oop"

	#
	# Squelch
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_SQUELCH
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_SQUELCH -label [mc Squelch] -underline 0 \
		-identifier M_SQUELCH

	set entries {}
	set i 1
	foreach {page label} $Priv(page) {
		lappend entries [list -type radiobutton -label $label \
			-variable ::NSSquelch($oop,radio,page) -value $page \
			-accelerator $i -identifier E_PAGE_$i]
		bind $win <KeyPress-$i> "NSSquelch::SetPage $oop $page"
		incr i
	}
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc Find...] \
		-accelerator f -underline 0 -identifier E_FIND]
	lappend entries [list -type command -label [mc "Find Again"] \
		-accelerator g -underline 6 -identifier E_FIND_AGAIN]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc "Save As Default"] \
		-underline 0 -identifier E_SAVE]
	lappend entries [list -type command -label [mc "Load Default"] \
		-underline 0 -identifier E_LOAD]
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc Close] \
		-underline 0 -accelerator $mod+W -identifier E_CLOSE]

	NSMenu::MenuInsertEntries $mbar -end MENU_SQUELCH $entries

	#
	# Mode
	#

	NSObject::New NSMenu $mbar -tearoff 0 -identifier MENU_MODE
	NSMenu::MenuInsertEntry $mbar -end MENUBAR -type cascade \
		-menu MENU_MODE -label [mc Mode] -underline 0 \
		-identifier M_MODE

	set entries {}
	lappend data \
		None NONE none n \
		Cursed CURSED cursed c \
		Average AVERAGE average v \
		Good GOOD good g \
		All ALL all a
	foreach {label ident mode key} $data {
		lappend entries [list -type radiobutton -label [mc $label] \
			-variable ::NSSquelch($oop,radio,mode) -value $mode \
			-accelerator $key -underline 0 -identifier E_MODE_$ident]
		bind $win <KeyPress-$key> "NSMenu::TryInvoke $mbar E_MODE_$ident"
		set MenuString(E_MODE_$ident) "Sets squelching mode to $mode."
	}
	lappend entries [list -type separator]
	lappend entries [list -type command -label [mc "Select All"] \
		-accelerator $mod+A -underline 0 -identifier E_SELECT_ALL]
	bind $win <Control-KeyPress-$key> "NSMenu::TryInvoke $mbar E_SELECT_ALL"

	NSMenu::MenuInsertEntries $mbar -end MENU_MODE $entries

	set MenuString(M_SQUELCH) \
		"Contains commands for ..."
	set MenuString(E_FIND) \
		"Searches for a member by name."
	set MenuString(E_FIND_AGAIN) \
		"Repeats the previous search."
	set MenuString(E_CLOSE) \
		"Closes the window."

	return
}

# NSSquelch::SetupMenus --
#
#	Prepare to post the menus.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SetupMenus {oop mbarID} {

	variable Priv

	lappend identList E_FIND E_FIND_AGAIN E_CLOSE
		
	set i 0
	foreach {page label} $Priv(page) {
		lappend identList E_PAGE_[incr i]
	}

	Info $oop radio,mode ""
	set numGroup [llength [Info $oop group,current]]
	set numMember [llength [Info $oop member,current]]
	if {($numGroup > 1) || ($numGroup == 1 && $numMember)} {
		lappend identList E_MODE_NONE E_MODE_ALL
		switch -- [Info $oop page] {
			page_armor -
			page_weapon {
				lappend identList E_MODE_CURSED E_MODE_AVERAGE E_MODE_GOOD
			}
			page_other {
			}
		}

		# If a single member is selected, then set the "Mode" radio entry
		if {$numMember == 1} {
			set group [lindex [Info $oop group,current] 0]
			set member [lindex [Info $oop member,current] 0]
			set k_idx [lindex [CallHook $oop member_list $group] $member]
			Info $oop radio,mode [angband k_info set $k_idx squelch]
		}
	}
	lappend identList E_SELECT_ALL

	NSMenu::MenuEnable $mbarID $identList

	[Info $oop win].statusBar cover show

	return
}

# NSSquelch::MenuSelect --
#
#	Displays a help string associated with a menu entry.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::MenuSelect {oop menuId index ident} {

	variable MenuString

	switch -glob -- $ident {
		E_PAGE_* {
			set desc "Displays this page."
		}
		
		default {
			if {[info exists MenuString($ident)]} {
				set desc $MenuString($ident)
			} else {
				set menu [NSMenu::Info $menuId menu]
				set desc [$menu entrycget $index -label]
			}
		}
	}

	[Info $oop win].statusBar cover set $desc
	if {![string length $desc]} {
		if {$menuId == [Info $oop mbarId]} {
			[Info $oop win].statusBar cover hide
		}
	}

	return
}

# NSSquelch::MenuInvoke --
#
#	Called when a menu entry is invoked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::MenuInvoke {oop menuId ident} {

	variable Priv

	switch -glob -- $ident {
		E_PAGE_* {
			set page [Info $oop radio,page]
			SetPage $oop $page
		}
		E_FIND {Find $oop 0}
		E_FIND_AGAIN {Find $oop 1}
		E_SAVE {}
		E_LOAD {}
		E_CLOSE {Close $oop}

		E_MODE_* {
			set mode [Info $oop radio,mode]
			set selGroup [Info $oop group,current]
			if {[llength $selGroup] == 1} {
				foreach row [Info $oop member,current] {
					set k_idx [lindex [Info $oop member,match] $row]
					angband k_info set $k_idx squelch $mode
					UpdateList $oop member $row squelch -text $mode
				}
			} else {
				foreach row $selGroup {
					foreach k_idx [CallHook $oop member_list $row] {
						angband k_info set $k_idx squelch $mode
					}
				}
			}
		}

		E_SELECT_ALL {
			set selGroup [Info $oop group,current]
			if {[llength $selGroup] == 1} {
				set canvistId [Info $oop member,canvistId]
				NSCanvist::UpdateSelection $canvistId all {}
			} else {
				set canvistId [Info $oop group,canvistId]
				NSCanvist::UpdateSelection $canvistId all {}
			}
		}
	}

	return
}

# NSSquelch::DisplayCmd --
#
#	Called by NSWindowManager::Display().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::DisplayCmd {oop message first args} {

	variable Priv
	
	switch -- $message {
		preDisplay {
			SetPage $oop [Info $oop page]
		}
		postDisplay {
		}
		postWithdraw {
			StateRemember $oop
			NSCanvist::DeleteAll [Info $oop group,canvistId]
			NSCanvist::DeleteAll [Info $oop member,canvistId]
		}
	}

	return
}

# NSSquelch::SetupCmd --
#
#	Called by NSWindowManager::Setup().
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SetupCmd {oop} {

	return
}

# NSSquelch::Close --
#
#	Description.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::Close {oop} {

	NSWindowManager::Undisplay squelch

	return
}

# NSSquelch::StateRemember --
#
#	Remember the group/member selections.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::StateRemember {oop} {

	if {!$NSWindowManager::Priv(squelch,setup)} return

	set page [Info $oop page]

	# Because we are clearing the lists here, and don't want to
	# upset the user, I save the selected group/member so it can
	# be restored in StateRestore() below.
	Info $oop group,$page [Info $oop group,current]
	Info $oop member,$page [Info $oop member,current]

	return
}

# NSSquelch::StateRestore --
#
#	Restore the group/member selections.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::StateRestore {oop} {

	set page [Info $oop page]
	
	# Restore the selected group
	set current [Info $oop group,$page]
	set canvistId [Info $oop group,canvistId]
	if {$current != {}} {
		NSCanvist::UpdateSelection $canvistId $current {}
		NSCanvist::See $canvistId [lindex $current 0]
	}

	# Restore the selected member
	set current [Info $oop member,$page]
	set canvistId [Info $oop member,canvistId]
	if {$current != {}} {
		NSCanvist::UpdateSelection $canvistId $current {}
		NSCanvist::See $canvistId [lindex $current 0]
	}

	return
}

# NSSquelch::CallHook --
#
#	Call the hook.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::CallHook {oop message args} {

	return [eval [Info $oop page] $oop $message $args]
}

# NSSquelch::SetPage --
#
#	Set the page.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SetPage {oop page} {

	variable Priv

	if {[string compare $page [Info $oop page]]} {
		set again 0
	} else {
		set again 1
	}

	if {!$again} {
		StateRemember $oop
	}

	# Remember the page
	Info $oop page $page

	# Set the group list
	SetList_Group $oop

	# Radiobutton menu entries
	Info $oop radio,page $page

	set tabsId [Info $oop tabsId]
	set current [NSTabs::Info $tabsId current]
	set tabId [NSTabs::GetNthId $tabsId [expr {[lsearch -exact $Priv(page) $page] / 2}]]
	if {$tabId != $current} {
		NSTabs::Smaller $tabsId $current
		NSTabs::Bigger $tabsId $tabId
		NSTabs::Info $tabsId current $tabId
	}

	StateRestore $oop
	
	return
}

# NSSquelch::SelectionChanged_Group --
#
#	When a "group" list item is selected, display members in the
#	"member" list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SelectionChanged_Group {oop canvistId select deselect} {

	set selection [NSCanvist::Selection $canvistId]
	set count [llength $selection]

	# Remember the selected groups
	Info $oop group,current $selection

	if {$count == 1} {

		# Get the row
		set row [lindex $selection 0]
	
		# Display members in that group
		SetList_Member $oop $row

	# Not 1 group selected, clear the member list
	} else {
		set canvistId [Info $oop member,canvistId]
		NSCanvist::DeleteAll $canvistId
	}

	return
}

# NSSquelch::SelectionChanged_Member --
#
#	Do something when a "member" list item is selected.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SelectionChanged_Member {oop canvistId select deselect} {

	set win [Info $oop win]

	set selection [NSCanvist::Selection $canvistId]
	set count [llength $selection]

	Info $oop member,current $selection

	# One item is selected
	if {$count == 1} {
	
		# Get the row
		set row [lindex $selection 0]
	
		# Call the hook to do other stuff (recall, etc)
		CallHook $oop select_member $row
		
		# Done
		return

	} else {

		$win.statusBar itemconfigure t2 -text ""
	}

	return
}

# NSSquelch::SetList_Group --
#
#	Set the list of groups.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SetList_Group {oop} {

	set win [Info $oop win]
	set canvistId [Info $oop group,canvistId]

	# Feedback
	StatusBar $oop "Displaying..." 0
	update idletasks

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	# Call hook to set the group list
	CallHook $oop set_list_group

	Info $oop group,current {}

	# Hack -- Clear the "member" list
	NSCanvist::DeleteAll [Info $oop member,canvistId]

	# Feedback
	StatusBar $oop "Done." 1

	return
}

# NSSquelch::SetList_Member --
#
#	Set the list of members.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::SetList_Member {oop group} {

	set canvistId [Info $oop member,canvistId]

	# Clear the list
	NSCanvist::DeleteAll $canvistId

	Info $oop member,current {}

	# Call hook to set the member list
	CallHook $oop set_list_member $group

	return
}

# NSSquelch::UpdateList --
#
#	Configure a canvas item on the row of a list.
#	When the user assigns an icon to a member, we want to update
#	the member list, and sometimes the group list too.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::UpdateList {oop which row tag args} {

	set canvistId [Info $oop $which,canvistId]
	set canvas [Info $oop $which,canvas]
	
	set rowTag [lindex [NSCanvist::Info $canvistId rowTags] $row]
	set itemIdList [$canvas find withtag $rowTag]
	set itemId [FindItemByTag $canvas $itemIdList $tag]
	eval $canvas itemconfigure $itemId $args

	return
}

# NSSquelch::CalcGroupListWidth --
#
#	Returns the desired width of the group list.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::CalcGroupListWidth {oop} {

	variable Priv

	# Get the desired font
	set font [Value font,knowledge]

	# Minimum desired width of the group list
	set maxWidth 100

	# Check each hook
	foreach {page title} $Priv(page) {

		# Check each name
		foreach name [$page $oop group_names] {

			# Calculate the width in pixels
			set width [font measure $font $name]
			
			# Remember the maximum width
			if {$width > $maxWidth} {
				set maxWidth $width
			}
		}
	}

	return [expr {[icon size] + 8 + $maxWidth + 20 + 4}]
}

# NSSquelch::Find --
#
#	Simple search routine to look for a member by name.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::Find {oop again} {

	variable Priv

	# Repeat the last search
	if {$again && [string length $Priv(find,string)]} {
		set string $Priv(find,string)

	# Enter a string to find, start from the beginning
	} else {

		# Ask the user for a name
		set string [NSUtils::StringBox -title [mc Find] \
			-initial $Priv(find,string) -prompt [mc find-prompt] \
			-buttons [list [mc Find] [mc Cancel]] -parent [Info $oop win]]

		# User cancelled
		if {![string length $string]} return

		# Clean up after the dialog, give message
		StatusBar $oop "Searching..." 1
		update

		# Reset search parameters
		set Priv(find,string) $string
	}

	# Default to searching from the beginning
	set group 0
	set member 0

	# Search in selected group, if any
	if {!$Priv(find,fromStart)} {	
		set groupCurr [Info $oop group,current]
		set memberCurr [Info $oop member,current]
		if {$groupCurr != {}} {
			set group [lindex $groupCurr 0]
		}
		if {$memberCurr != {}} {
			set member [expr {[lindex $memberCurr 0] + 1}]
		}
	}

	set max [NSCanvist::Info [Info $oop group,canvistId] count]

	# Compare lowercase
	set string [string tolower $string]

	# Check each group
	for {set i $group} {$i < $max} {incr i} {

		# Get a list of member indexes
		set match [lrange [CallHook $oop member_list $i] $member end]

		# Check each member index
		foreach index $match {
	
			# Get the member name
			set name2 [CallHook $oop member_name $i $index]

			# Compare lowercase
			set name2 [string tolower $name2]

			# Found a match
			if {[string first $string $name2] != -1} {
	
				set canvistId [Info $oop group,canvistId]
	
				# The new group is not displayed
				if {![NSCanvist::IsRowSelected $canvistId $i]} {

					# Clear the current selection
					NSCanvist::RemoveSelection $canvistId
					
					# Select the new group. As a side effect, the
					# SetList_Member() command is called to display
					# the monsters in the group.
					NSCanvist::UpdateSelection $canvistId $i {}
					NSCanvist::See $canvistId $i

				# The new group is already selected
				} else {
				}
	
				set canvistId [Info $oop member,canvistId]
	
				# Select the matching member, deselecting all others
				NSCanvist::UpdateSelection $canvistId $member all
				NSCanvist::See $canvistId $member
	
				# Don't search from start next time
				set Priv(find,fromStart) 0

				# Clear "Searching..." message
				StatusBar $oop "" 0

				# Done
				return
			}

			# Next row
			incr member
		}

		# First row
		set member 0
	}

	# If we didn't search from the start, then wrap around
	if {!$Priv(find,fromStart)} {
		set Priv(find,fromStart) 1
		Find $oop 1
		return
	}

	StatusBar $oop "No match for \"$string\"." 1

	return
}

# NSSquelch::StatusBar --
#
#	Display text in the status bar, perhaps clearing it later.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::StatusBar {oop text zap} {

	set win [Info $oop win]
	set label [$win.statusBar itemcget t1 -label]
	$label configure -text $text
	if {$zap} {
		NSUtils::ZapLabel $label
	}

	return
}

# NSSquelch::NewItemCmd --
#
#	Called by NSCanvist::InsertItem() to create a list row.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::NewItemCmd {oop canvistId y args} {

	array set config [list -icon "default 0" -text Unknown -squelch ""]
	array set config $args

	set c [NSCanvist::Info $canvistId canvas]
	set lineHeight [NSCanvist::Info $canvistId rowHgt]
	set font [Value font,knowledge]

	set fh [font metrics $font -linespace]
	set diff [expr {int([expr {($lineHeight - $fh) / 2}])}]

	# Get the width of the canvas
	set canvasWidth [winfo width $c]

	# Selection rectangle around everything
	lappend itemIdList [$c create rectangle 2 [expr {$y + 2}] \
		[expr {$canvasWidth - 2}] [expr {$y + $lineHeight - 2}] \
		-fill {} -outline {} -tags selrect -width 2.0]

	# Image
	set iw [icon size]
	set ih [icon size]
	set wid [expr {[icon size] + 8}]
	set xdiff [expr {int([expr {($wid - $iw) / 2}])}]
	set ydiff [expr {int([expr {($lineHeight - $ih) / 2}])}]
	lappend itemIdList [$c create widget $xdiff [expr {$y + $ydiff}] \
		-assign $config(-icon) -tags icon]

	# Text
	set fill White
	set textLeft [expr {$wid + 1}]
	lappend itemIdList [$c create text $textLeft [expr {$y + $diff}] \
		-text $config(-text) -anchor nw -font $font -fill $fill -tags text]

	# Squelch setting
	if {[string length $config(-squelch)]} {
		lappend itemIdList [$c create text [expr {($canvasWidth - 1) - 4}] \
			 [expr {$y + $diff}] -text $config(-squelch) -anchor ne \
			 -font $font -fill $fill -tags squelch]
	}

	return $itemIdList
}

# NSSquelch::HighlightItemCmd --
#
#	Called by NSCanvist::Select() to highlight a row.
#
# Arguments:
#	oop					OOP ID. See above.
#	canvistId					OOP ID of NSCanvist object.
#	state					1 or 0 highlight state.
#	args					List of canvas item ids
#
# Results:
#	What happened.

proc NSSquelch::HighlightItemCmd {oop canvistId state args} {

	set canvas [NSCanvist::Info $canvistId canvas]
	set itemIdList $args

	set idRect [FindItemByTag $canvas $itemIdList selrect]

	if {[NSUtils::HasFocus $canvas]} {
		set fill [Value listHilite]
	} else {
		set fill [Value listInactive]
	}

	if {$state} {
		$canvas itemconfigure $idRect -outline $fill

	} else {
		$canvas itemconfigure $idRect -outline {}
	}

	return
}

# NSSquelch::Configure --
#
#	Called as a <Configure> event script. Resizes the selection rectangles
#	so they fit properly.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::Configure {oop canvas} {

	# Get the width of the canvas
	set canvasWidth [winfo width $canvas]

	foreach itemId [$canvas find withtag selrect] {
		set coords [$canvas coords $itemId]
		set right [expr {$canvasWidth - 2}]
		set coords [lreplace $coords 2 2 $right]
		eval $canvas coords $itemId $coords
	}

	foreach itemId [$canvas find withtag squelch] {
		set coords [$canvas coords $itemId]
		set right [expr {($canvasWidth - 1) - 4}]
		set coords [lreplace $coords 0 0 $right]
		eval $canvas coords $itemId $coords
	}

	return
}

# NSSquelch::InvokeTab --
#
#	Called when a tab is clicked.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::InvokeTab {oop tabsId tabId} {

	variable Priv

	set index [lsearch -exact [NSTabs::Info $tabsId id] $tabId]
	SetPage $oop [lindex $Priv(page) [expr {$index * 2}]]

	return
}

# NSSquelch::ValueChanged_font_knowledge --
#
#	Called when the font,knowledge value changes.
#	Updates the Assign Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::ValueChanged_font_knowledge {oop} {

	# Get the desired font
	set font [Value font,knowledge]

	# Calculate row height
	set rowHgt [font metrics $font -linespace]
	if {[icon size] > $rowHgt} {
		set rowHgt [icon size]
	}
	incr rowHgt 8

	# Set row height of group list
	set canvistId [Info $oop group,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	# Set the width of the group list
	set oldWidth [winfo width $canvas]
	set newWidth [CalcGroupListWidth $oop]
	$canvas configure -width $newWidth

	# Set the scrollregion to prevent horizontal scrolling
	scan [$canvas cget -scrollregion] "%s %s %s %s" x1 y1 x2 y2
	$canvas configure -scrollregion "$x1 $y1 $newWidth $y2"

	# Hack -- Resize the toplevel so the member list is not resized
	set diff [expr {$newWidth - $oldWidth}]
	if {$diff} {
		set win [Info $oop win]
		set newWidth [expr {[winfo width $win] + $diff}]
		NSToplevel::SetTotalWidth $win $newWidth
	}

	# Set row height of member list
	set canvistId [Info $oop member,canvistId]
	set canvas [NSCanvist::Info $canvistId canvas]
	NSCanvist::Info $canvistId rowHgt $rowHgt
	$canvas configure -yscrollincrement $rowHgt

	return
}

# NSSquelch::ValueChanged_listBG --
#
#	Called when the listBG value changes.
#	Updates the Assign Window.
#
# Arguments:
#	arg1					about arg1
#
# Results:
#	What happened.

proc NSSquelch::ValueChanged_listBG {oop} {

	set color [Value listBG]
	
	[Info $oop group,canvas] configure -background $color
	[Info $oop member,canvas] configure -background $color

	return
}

proc NSSquelch::page_armor {oop message args} {

	variable Priv

	switch -- $message {

		set_list_group {

			set canvistId [Info $oop group,canvistId]

			foreach {title findSpec} $Priv(group,armor) {

				# Find the last monster in the group
				set match [eval angband k_info find -limit 1 \
					-backwards $findSpec]

				# Get the assignment. This resolves alternates.
				set icon [angband k_info info [lindex $match 0] icon]
		
				# Add this group to the list
				NSCanvist::Insert $canvistId end -icon $icon -text [mc $title]
			}
		}

		set_list_member {

			set canvistId [Info $oop member,canvistId]

			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,armor) [expr {$group * 2 + 1}]]

			# Get a list of monsters in the group
			set match [eval angband k_info find $findSpec]

			# Collect info for each row
			set itemList {}

			# Add each match to the list
			foreach index $match {
		
				# Get the assignment. This resolves alternates.
				set icon [angband k_info info $index icon]

				# Hack -- object_desc
				set name [angband k_info info $index object_desc]

				set squelch [angband k_info set $index squelch]

				# Collect info for each row
				lappend itemList [list -icon $icon -text $name \
					-squelch $squelch]
			}

			# Add each row to the list
			NSCanvist::InsertMany $canvistId end $itemList

			# Keep a list of matching indexes
			Info $oop member,match $match
		}

		select_member {
			set row [lindex $args 0]
			set index [lindex [Info $oop member,match] $row]
[Info $oop win].statusBar itemconfigure t2 -text $index
			NSRecall::RecallObjectKind $index
		}

		group_names {
			set names {}
			foreach {title findSpec} $Priv(group,armor) {
				lappend names [mc $title]
			}
			return $names
		}

		group_list {
			set index -1
			set result {}
			foreach {title findSpec} $Priv(group,armor) {
				lappend result [incr index]
			}
			return $result
		}

		member_name {
			set group [lindex $args 0]
			set index [lindex $args 1]
			return [angband k_info info $index object_desc]
		}

		member_list {
			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,armor) [expr {$group * 2 + 1}]]
			return [eval angband k_info find $findSpec]
		}
	}

	return
}

proc NSSquelch::page_weapon {oop message args} {

	variable Priv

	switch -- $message {

		set_list_group {

			set canvistId [Info $oop group,canvistId]

			foreach {title findSpec} $Priv(group,weapon) {

				# Find the last monster in the group
				set match [eval angband k_info find -limit 1 \
					-backwards $findSpec]

				# Get the assignment. This resolves alternates.
				set icon [angband k_info info [lindex $match 0] icon]
		
				# Add this group to the list
				NSCanvist::Insert $canvistId end -icon $icon -text [mc $title]
			}
		}

		set_list_member {

			set canvistId [Info $oop member,canvistId]

			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,weapon) [expr {$group * 2 + 1}]]

			# Get a list of monsters in the group
			set match [eval angband k_info find $findSpec]

			# Collect info for each row
			set itemList {}

			# Add each match to the list
			foreach index $match {
		
				# Get the assignment. This resolves alternates.
				set icon [angband k_info info $index icon]

				# Hack -- object_desc
				set name [angband k_info info $index object_desc]

				set squelch [angband k_info set $index squelch]

				# Collect info for each row
				lappend itemList [list -icon $icon -text $name \
					-squelch $squelch]
			}

			# Add each row to the list
			NSCanvist::InsertMany $canvistId end $itemList

			# Keep a list of matching indexes
			Info $oop member,match $match
		}

		select_member {
			set row [lindex $args 0]
			set index [lindex [Info $oop member,match] $row]
			NSRecall::RecallObjectKind $index
		}

		group_names {
			set names {}
			foreach {title findSpec} $Priv(group,weapon) {
				lappend names [mc $title]
			}
			return $names
		}

		group_list {
			set index -1
			set result {}
			foreach {title findSpec} $Priv(group,weapon) {
				lappend result [incr index]
			}
			return $result
		}

		member_name {
			set group [lindex $args 0]
			set index [lindex $args 1]
			return [angband k_info info $index object_desc]
		}

		member_list {
			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,weapon) [expr {$group * 2 + 1}]]
			return [eval angband k_info find $findSpec]
		}
	}

	return
}

proc NSSquelch::page_other {oop message args} {

	variable Priv

	switch -- $message {

		set_list_group {

			set canvistId [Info $oop group,canvistId]

			foreach {title findSpec} $Priv(group,other) {

				# Find the last monster in the group
				set match [eval angband k_info find -limit 1 \
					-backwards $findSpec]

				# Get the assignment. This resolves alternates.
				set icon [angband k_info info [lindex $match 0] icon]
		
				# Add this group to the list
				NSCanvist::Insert $canvistId end -icon $icon -text [mc $title]
			}
		}

		set_list_member {

			set canvistId [Info $oop member,canvistId]

			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,other) [expr {$group * 2 + 1}]]

			# Keep a list of matching indexes
			set match {}

			# Collect info for each row
			set itemList {}

			# Add each match to the list
			foreach index [eval angband k_info find $findSpec] {

				# Skip INSTA_ART objects
				if {[angband k_info info $index insta_art]} continue

				# Get the assignment. This resolves alternates.
				set icon [angband k_info info $index icon]

				# Hack -- object_desc
				set name [angband k_info info $index object_desc]

				set squelch [angband k_info set $index squelch]

				# Collect info for each row
				lappend itemList [list -icon $icon -text $name \
					-squelch $squelch]

				# Keep a list of matching indexes
				lappend match $index
			}

			# Add each row to the list
			NSCanvist::InsertMany $canvistId end $itemList

			# Keep a list of matching indexes
			Info $oop member,match $match
		}

		select_member {
			set row [lindex $args 0]
			set index [lindex [Info $oop member,match] $row]
			NSRecall::RecallObjectKind $index
		}

		group_names {
			set names {}
			foreach {title findSpec} $Priv(group,other) {
				lappend names [mc $title]
			}
			return $names
		}

		group_list {
			set index -1
			set result {}
			foreach {title findSpec} $Priv(group,other) {
				lappend result [incr index]
			}
			return $result
		}

		member_name {
			set group [lindex $args 0]
			set index [lindex $args 1]
			return [angband k_info info $index object_desc]
		}

		member_list {
			set group [lindex $args 0]
			set findSpec [lindex $Priv(group,other) [expr {$group * 2 + 1}]]
			foreach index [eval angband k_info find $findSpec] {

				# Skip INSTA_ART objects
				if {[angband k_info info $index insta_art]} continue

				lappend result $index
			}
			return $result
		}
	}

	return
}

